#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""A specialized verifier for the NN4sys benchmark."""
import time
import pdb
import arguments

from load_model import inference_onnx
from auto_LiRPA import BoundedModule, BoundedTensor
from auto_LiRPA.perturbations import *
from auto_LiRPA.bound_ops import *
from collections import deque
import numpy as np


@torch.no_grad()
def compute_bounds(model_v1, model_v2, input_l, input_u, x=None, eps=None):
    if x is None or eps is None:
        x = (input_l + input_u) / 2
        eps = (input_u - input_l) / 2
    ptb = PerturbationLpNorm(norm=np.inf, x_L=input_l, x_U=input_u)
    x = BoundedTensor(x, ptb).to(x.device)
    l, u = model_v1.compute_bounds(x=(x,), method='forward')
    # check that there won't be two positions with positive values after the first 1e10 layer
    margin = ( (u<=0).int() * 1e20 + (u>0).int() * l )[:, :-1].min(dim=-1).values
    safe_numerical_1 = (u[:, :-1].max(dim=-1).values - margin * 1e10 < 0)
    # check that there will be at least one positive with a positive value after the first 1e10 layer
    safe_numerical_2 = (l[:, 0] > 1e-8)

    select = (u[:, :-1] > 0).int()
    select[:, :-1] *= ( u[:,:-2] - l[:, 1:-1]*1e10 >= 1e-8).int() # 1e-8 is a tolerance term
    l, u = model_v2.compute_bounds(x=(x,), method='forward')
    l_final = torch.min(select * l + (1 - select) * (1e10), dim=-1).values
    u_final = torch.max(select * u + (1 - select) * (-1e10), dim=-1).values

    safe_numerical = (safe_numerical_1 * safe_numerical_2).int()
    l_final = l_final * safe_numerical + l_final.clamp(max=0) * (1 - safe_numerical)

    return l_final, u_final

@torch.no_grad()
def verify(model_ori, model_v1, model_v2, data_ub=None, data_lb=None, spec=None):
    device = arguments.Config["general"]["device"]

    data_lb, data_ub = data_lb.to(device), data_ub.to(device)
    # use forward mode
    # y = model(x) # NOTE currently must call a clean forward before computing forward bounds
    queue = deque() # a queue of input segments to be verified
    queue.append((data_lb, data_ub))
    base = None
    res_l, res_u = None, None
    max_queue_size = int(1e6) # TODO confirm it
    input_length = float(data_ub - data_lb)
    first = True
    batch_size = 32
    split_batch = 16
    while len(queue) > 0:
        print('\rqueue size', len(queue), 'input length', input_length, end='')
        # TODO timeout
        # TODO batching, we can draw len(args.batch_size, len(queue)) examples from the queue directly.
        # seg_l, seg_u: input segment
        seg_l, seg_u = [], []
        while len(queue) > 0 and len(seg_l) < batch_size:
            seg_l_, seg_u_ = queue.pop()
            seg_l.append(seg_l_)
            seg_u.append(seg_u_)
        seg_l = torch.cat(seg_l).view(len(seg_l), -1)
        seg_u = torch.cat(seg_u).view(len(seg_u), -1)
        input_length -= (seg_u - seg_l).sum()
        eps = (seg_u - seg_l) / 2
        x = (seg_u + seg_l) / 2
        l_final, u_final = compute_bounds(model_v1, model_v2, seg_l, seg_u, x=x, eps=eps)
        for k in range(seg_l.size(0)):
            l, u = l_final[k].item(), u_final[k].item()
            # if l == 0 or u == 0:
            #     print(f'\nWARNING: numerical error detected, seg_l={seg_l[k].item()}, seg_u={seg_u[k].item()}, l={l}, u={u}')
            #     pdb.set_trace()
            if base is None:
                base = (l, u)
            verified = l >= spec[2] if spec[3] == 'lower' else u <= spec[2]
            if verified:
                if res_l is None:
                    res_l, res_u = l, u
                else:
                    res_l = min(res_l, l)
                    res_u = max(res_u, u)
            else:
                has_unstable = False
                for r in model_v1.relus + model_v2.relus:
                    has_unstable = (has_unstable or
                        torch.logical_and(r.inputs[0].lower[k] < 0, r.inputs[0].upper[k] > 0).any())
                if has_unstable:
                    eps_ = eps[k].item()
                    # first time to split
                    if first:
                        for i in range(split_batch):
                            queue.append((x[k:(k+1)] - (eps_/split_batch) * (i+1),
                                x[k:(k+1)] - (eps_/split_batch) * i))
                            queue.append((x[k:(k+1)] + (eps_/split_batch) * i,
                                x[k:(k+1)] + (eps_/split_batch) * (i+1)))
                        first = False
                    else:
                        # NOTE: simply using (seg_l, x) and (x, seg_u) would have
                        # numerical issues and endless branching
                        queue.append((x[k:(k+1)]-eps_, x[k:(k+1)]))
                        queue.append((x[k:(k+1)], x[k:(k+1)]+eps_))
                    input_length += eps_ * 2
                    if len(queue) > max_queue_size:
                        # Too many. Give up
                        return base
                else:
                    print(f'Adversarial example exists: l={l}, u={u}')
                    pred_l, pred_u = model_ori[-1](seg_l[k:(k+1)]), model_ori[-1](seg_u[k:(k+1)])
                    print(f'Model pred: {seg_l[k].item():.10g}->{pred_l.item()}, {seg_u[k].item():.10g}->{pred_u.item()}')
                    unsafe = False
                    if spec[3] == 'lower':
                        if pred_l < spec[2] or pred_u < spec[2]: unsafe = True
                    else:
                        if pred_l > spec[2] or pred_u > spec[2]: unsafe = True
                    if unsafe:
                        return base[0], base[1], 'unsafe'
                    else:
                        return base[0], base[1]
    print()
    print('bab_single_input output', res_l, res_u)
    return res_l, res_u

""" Attack nn4sys:
Let y be the output of model_v1.
If there exists y[i], y[i+1], y[i+2], such that
y[i]-y[i+1]*1e10>0, y[i]-y[i+1]*2e10<0, y[i+2]=0,
the output of the original model will be 0 due to the numerical issue.
"""
@torch.no_grad()
def attack_nn4sys(model_ori, model_v1, spec, onnx_path, device):
    def query(x):
        x = torch.tensor(x, dtype=torch.float64, device=device).view(1, 1)
        return F.relu(model_v1(x)[0])
    out_l = query(spec[0])
    out_u = query(spec[1])
    idx = -1
    for i in range(1, out_l.numel() - 2):
        if out_u[i + 1] == 0 and out_l[i] == 0 and out_u[i] > 0:
            idx = i
            break
    if idx == -1:
        return None
    l = spec[0]
    r = spec[1]
    while l + 1e-20 < r:
        m = (l + r) / 2
        out = query(m)
        if out[idx].item() == 0:
            l = m
        elif (out[idx-1]-1e10*out[idx]) < 0:
            r = m
        elif (out[idx-1]-2e10*out[idx]) > 0:
            l = m
        else:
            # TODO confirm that onnxruntime gives same results
            x = torch.tensor(m, dtype=torch.float64, device=device).view(1, 1)
            out_model_ori = model_ori[-1](x)
            print('attack result', m)
            print('model_ori output', out_model_ori)
            out_onnx = inference_onnx(onnx_path, x[0].cpu().numpy())
            print('onnx output', out_onnx)
            if out_model_ori == 0:
                return m
            else:
                return None
    return None

@torch.no_grad()
def nn4sys_verification(model_ori, vnnlib, onnx_path):
    start_time = time.time()
    all_verified = True
    unsafe = False

    device = arguments.Config["general"]["device"]
    for i in range(3):
        model_ori[i].to(device=device)
    model_v1 = BoundedModule(
        model_ori[0],
        torch.empty((2, 1), dtype=torch.float64, device=device),
        device=device)
    model_v2 = BoundedModule(
        model_ori[1],
        torch.empty((2, 1), dtype=torch.float64, device=device),
        device=device)

    # When there are more than 1 properties, we batch all properties to be verified, and check all of them once in a big batch. Many of them can be verified directly without splitting.
    # Then, we only split input those properties that cannot be verified directly. This allows us to solve many instances instantly.
    if len(vnnlib) > 2:
        data_min = torch.tensor([spec[0] for spec in vnnlib], dtype=torch.float64).view(-1, 1).to(device)
        data_max = torch.tensor([spec[1] for spec in vnnlib], dtype=torch.float64).view(-1, 1).to(device)
        l_pre, u_pre = compute_bounds(model_v1, model_v2, data_min, data_max)

    for k, spec in enumerate(vnnlib):
        print(f'Verifying spec with {spec[0]} <= x <= {spec[1]} with output {spec[3]} {spec[2]}')
        if spec[-1] == 'lower' and spec[2] > 0:
            attack_x = attack_nn4sys(model_ori, model_v1, spec, onnx_path, device)
            print('attack', attack_x)
            if attack_x is not None:
                unsafe = True
                break

        if len(vnnlib) > 2:
            if spec[-1] == 'upper':
                verified = u_pre[k] <= spec[2]
            else:
                verified = l_pre[k] >= spec[2]
            if verified:
                print(f'Verified from precomputed bounds {l_pre[k].item():.6f} {u_pre[k].item():.6f}')
                continue

        data_min = torch.tensor(spec[0], dtype=torch.float64).view(1, 1)
        data_max = torch.tensor(spec[1], dtype=torch.float64).view(1, 1)
        res = verify(model_ori, model_v1, model_v2, data_ub=data_max, data_lb=data_min, spec=spec)
        if len(res) == 3:
            assert res[2] == 'unsafe'
            unsafe = True
            break
        else:
            l, u = res
            if False:
                # search for check
                from tqdm import tqdm
                step = 0.1
                batch_size = 1024
                x = torch.arange(spec[0], spec[1], step, dtype=torch.float64, device=device)
                print(f'Brute force check, {spec[0]}~{spec[1]}, {x.numel()} points')
                bl, bu = 1e10, -1e10
                for i in tqdm(range((x.numel() + batch_size - 1) // batch_size)):
                    x_ = x[i*batch_size:(i+1)*batch_size].view(-1, 1)
                    out = model_ori[-1](x_)
                    try:
                        assert l <= out.min()
                        assert u >= out.max()
                    except AssertionError:
                        pdb.set_trace()
                    bl = min(bl, out.min().item())
                    bu = max(bu, out.max().item())
                print(f'Verification {l} {u}')
                print(f'Brute force {bl} {bu}')
                # pdb.set_trace()

        print('spec', spec)
        print('verified lower', l)
        print('verified upper', u)
        if spec[3] == 'lower':
            verified = l >= spec[2]
        else:
            verified = u <= spec[2]
        if not verified:
            all_verified = False
            break

    if unsafe:
        res = 'unsafe'
    elif all_verified and time.time() - start_time <= arguments.Config["bab"]["timeout"] - 5:
        res = 'safe'
    else:
        res = 'unknown'

    print('Result', res)
    print('Time', time.time() - start_time)

    return res



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import torch
import arguments

from prune import prune_alphas

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from beta_CROWN_solver import LiRPANet


def copy_alpha(self: 'LiRPANet', reference_alphas, num_targets,
               target_batch_size=None, now_batch=None, interm_bounds=None,
               batch_size=None):
    # alpha manipulation, since after init_alpha all things are copied
    # from alpha-CROWN and these alphas may have wrong shape
    opt_interm_bounds = arguments.Config['solver']['beta-crown']['enable_opt_interm_bounds']
    final_name = self.net.final_node().name
    for m in self.net.optimizable_activations:
        keys = list(m.alpha.keys())
        # when fixed intermediate bounds are available, since intermediate betas
        # are not used anymore because we use fixed intermediate bounds later,
        # we can delete these intermediate betas to save space
        if interm_bounds is not None and not opt_interm_bounds:
            for k in keys:
                if k not in self.alpha_start_nodes:
                    del m.alpha[k]
        if final_name not in m.alpha:
            continue
        if (m.alpha[final_name].shape[1] != 1
                or m.alpha[final_name].shape[2] != batch_size):
            # shape mismatch detected
            # pick the first slice with shape [2, 1, 1, ...],
            # and repeat to [2, 1, batch_size, ...]
            repeat = [1 if i != 2 else batch_size
                    for i in range(m.alpha[final_name].dim())]
            m.alpha[final_name] = m.alpha[final_name][:, 0:1, 0:1].repeat(*repeat)

    if reference_alphas is None:
        return False

    # We already have alphas available
    all_alpha_initialized = True
    for m in self.net.optimizable_activations:
        for spec_name, alpha in m.alpha.items():
            # each alpha size is (2, spec, batch_size, *shape); batch size is 1.
            if not spec_name in reference_alphas[m.name]:
                continue
            reference_alpha = reference_alphas[m.name][spec_name]
            if spec_name == self.net.final_node().name:
                target_start = now_batch * target_batch_size
                target_end = min((now_batch + 1) * target_batch_size, num_targets)
                if alpha.size()[2] == target_end - target_start:
                    print(f'setting alpha for layer {m.name} '
                          f'start_node {spec_name} with alignment adjustment')
                    # The reference alpha has deleted the pred class itself,
                    # while our alpha keeps that
                    # now align these two
                    # note: this part actually implements the following
                    # TODO (extract alpha according to different label)
                    if reference_alpha.size()[1] > 1:
                        # didn't apply multiple x in incomplete_verifier
                        alpha.data = reference_alpha[:, target_start:target_end].reshape_as(alpha.data)
                    else:
                        # applied multiple x in incomplete_verifier
                        alpha.data = reference_alpha[:, :, target_start:target_end].reshape_as(alpha.data)
                else:
                    all_alpha_initialized = False
                    print(f'not setting layer {m.name} start_node {spec_name} '
                          'because shape mismatch '
                          f'({alpha.size()} != {reference_alpha.size()})')
            elif alpha.size() == reference_alpha.size():
                print(f"setting alpha for layer {m.name} start_node {spec_name}")
                alpha.data.copy_(reference_alpha)
            elif all([si == sj or ((d == 2) and sj == 1)
                      for d, (si, sj) in enumerate(
                          zip(alpha.size(), reference_alpha.size()))]):
                print(f'setting alpha for layer {m.name} start_node {spec_name} '
                      'with batch sample broadcasting')
                alpha.data.copy_(reference_alpha)
            else:
                # TODO extract alpha according to different label
                all_alpha_initialized = False
                print(f'not setting layer {m.name} start_node {spec_name} '
                      'because shape mismatch '
                      f'({alpha.size()} != {reference_alpha.size()})')

    return all_alpha_initialized


def get_alpha(self: 'LiRPANet', only_final=False, half=False, device=None):
    # alpha has size (2, spec, batch, *shape). When we save it,
    # we make batch dimension the first.
    # spec is some intermediate layer neurons, or output spec size.
    ret = {}
    for m in self.net.perturbed_optimizable_activations:
        ret[m.name] = {}
        for spec_name, alpha in m.alpha.items():
            if not only_final or spec_name in self.alpha_start_nodes:
                ret[m.name][spec_name] = self._transfer(alpha, device, half=half)
    return ret


def set_alpha(self: 'LiRPANet', d, set_all=False):
    alpha = d['alphas']
    if len(alpha) == 0:
        return

    for m in self.net.perturbed_optimizable_activations:
        for spec_name in list(m.alpha.keys()):
            if spec_name in alpha[m.name]:
                # Only setup the last layer alphas if no refinement is done.
                if spec_name in self.alpha_start_nodes or set_all:
                    m.alpha[spec_name] = alpha[m.name][spec_name]
                    # Duplicate for the second half of the batch.
                    m.alpha[spec_name] = m.alpha[spec_name].detach().requires_grad_()
            else:
                # This layer's alpha is not used.
                # For example, we can drop all intermediate layer alphas.
                del m.alpha[spec_name]


def add_batch_alpha(self, alphas, reference_alphas):
    batch_alphas = self.get_alpha()
    if arguments.Config['bab']['attack']['enabled']:
        # Save all alphas, which will be further refined in bab-attack.
        self.refined_alpha = reference_alphas
    # early alpha delete to save space
    # If we are not optimizing intermediate layer bounds, we do not need
    # to save all the intermediate alpha. We only keep the alpha for the
    # last layer.`
    new_batch_alphas = prune_alphas(batch_alphas, self.alpha_start_nodes)
    del batch_alphas
    batch_alphas = new_batch_alphas
    for k, v in batch_alphas.items():
        if k not in alphas:
            alphas[k] = {}
        for kk, vv in v.items():
            if kk not in alphas[k]:
                alphas[k][kk] = vv
            else:
                alphas[k][kk] = torch.cat([alphas[k][kk], vv], dim=2)



from functools import partial
import os
import torch
import torch.nn as nn
import numpy as np
from auto_LiRPA.bound_general import BoundedModule

from smac import Scenario, HyperparameterOptimizationFacade
from smac.utils.configspace import get_config_hash

from CTRAIN.eval.eval import eval_acc, eval_model, eval_complete_abcrown
from CTRAIN.model_wrappers.configs import get_config_space


class CTRAINWrapper(nn.Module):
    
    def __init__(self, model: nn.Module, eps:float, input_shape: tuple, train_eps_factor=1, lr=0.0005, optimizer_func=torch.optim.Adam, bound_opts=dict(conv_mode='patches', relu='adaptive'), device='cuda', checkpoint_save_path=None):
        super(CTRAINWrapper, self).__init__()
        model = model.to(device)
        
        original_train = model.training
        self.original_model = model
        self.eps = eps
        self.train_eps = eps * train_eps_factor
        if isinstance(device, torch.device):
            self.device = device
        else:
            if device in ['cuda', 'cpu', 'mps']:
                self.device = torch.device(device)
            else:
                print("Unknown device - falling back to device CPU!")
                self.device = torch.device('cpu')
        
        if len(input_shape) < 4:
            input_shape = [1, *input_shape]
        model.eval()
        example_input = torch.ones(input_shape, device=device)
        self.n_classes = len(model(example_input)[0])
        self.bound_opts = bound_opts
        self.bounded_model = BoundedModule(model=self.original_model, global_input=example_input, bound_opts=bound_opts, device=device)
        self.input_shape = input_shape
        
        self.optimizer_func = optimizer_func
        self.optimizer = optimizer_func(self.bounded_model.parameters(), lr=lr)
        
        self.epoch = 0
        
        if original_train:
            self.original_model.train()
            self.bounded_model.train()
        
        self.checkpoint_path = checkpoint_save_path
        if checkpoint_save_path is not None:
            os.makedirs(self.checkpoint_path, exist_ok=True)
    
    def train(self):
        self.original_model.train()
        self.bounded_model.train()
    
    def eval(self):
        self.original_model.train()
        self.bounded_model.train()
    
    def forward(self, x):
        return self.bounded_model(x)
    
    def evaluate(self, test_loader, test_samples=np.inf):
        eps_std = self.eps / test_loader.std if test_loader.normalised else torch.tensor(self.eps)
        eps_std = torch.reshape(eps_std, (*eps_std.shape, 1, 1))
        return eval_model(self.bounded_model, test_loader, n_classes=self.n_classes, eps=eps_std, test_samples=test_samples, method='ADAPTIVE', device=self.device)

    def evaluate_complete(self, test_loader, test_samples=np.inf, timeout=1000, no_cores=4, abcrown_batch_size=512):
        eps_std = self.eps / test_loader.std if test_loader.normalised else self.eps
        eps_std = torch.reshape(eps_std, (*eps_std.shape, 1, 1))
        std_acc = eval_acc(self.bounded_model, test_loader=test_loader, test_samples=test_samples)
        certified_acc, adv_acc = eval_complete_abcrown(
            model=self.bounded_model,
            eps_std=eps_std,
            data_loader=test_loader,
            n_classes=self.n_classes,
            input_shape=self.input_shape,
            test_samples=test_samples,
            timeout=timeout,
            no_cores=no_cores,
            abcrown_batch_size=abcrown_batch_size,
            device=self.device
        )
        return std_acc, certified_acc, adv_acc
    
    def state_dict(self):
        return self.bounded_model.state_dict()
    
    def load_state_dict(self, state_dict, strict = True):
        return self.bounded_model.load_state_dict(state_dict, strict)
    
    def parameters(self, recurse=True):
        return self.bounded_model.parameters(recurse=recurse)
    # TODO: Add onnx export/loading

    def resume_from_checkpoint(self, checkpoint_path:str, train_loader, val_loader=None):
        checkpoint = torch.load(checkpoint_path)
        model_state_dict = checkpoint['model_state_dict']
        self.load_state_dict(model_state_dict)
        self.epoch = checkpoint['epoch']
        optimizer_state_dict = checkpoint['optimizer_state_dict']
        self.optimizer.load_state_dict(optimizer_state_dict)
        self.train_model(train_loader, val_loader, start_epoch=self.epoch)

    
    
    def hpo(self, train_loader, val_loader, budget=5*24*60*60, defaults=dict(), eval_samples=1000, output_dir='./smac_hpo', include_nat_loss=True, include_adv_loss=True, include_cert_loss=True):
        os.makedirs(output_dir, exist_ok=True)
        if os.listdir(output_dir):
            assert False, 'Output directory for HPO is not empty!'
        
        os.makedirs(f'{output_dir}/nets', exist_ok=True)
        os.makedirs(f'{output_dir}/smac/', exist_ok=True)

        eps_std = self.eps / train_loader.std
        scenario = Scenario(
            configspace=get_config_space(self, self.num_epochs, eps_std, defaults=defaults),
            walltime_limit=budget,
            n_trials=np.inf,
            output_directory=f'{output_dir}/smac/',
            use_default_config=True if len(defaults.values()) > 0 else False
        )
        initial_design = HyperparameterOptimizationFacade.get_initial_design(scenario, n_configs_per_hyperparamter=1)
        smac = HyperparameterOptimizationFacade(
            scenario,
            partial(self._hpo_runner, epochs=self.num_epochs, train_loader=train_loader, val_loader=val_loader, cert_eval_samples=eval_samples, output_dir=output_dir, include_nat_loss=include_nat_loss, include_adv_loss=include_adv_loss, include_cert_loss=include_cert_loss),
            initial_design=initial_design,
            overwrite=True,
        )

        inc = smac.optimize()
        
        config_hash = get_config_hash(inc, 32)
        self.load_state_dict(torch.load(f'{output_dir}/nets/{config_hash}.pt'))

        return inc

    def _hpo_runner(self, config, seed, epochs, train_loader, val_loader, output_dir, cert_eval_samples=1000):
        raise NotImplementedError('HPO can only be run on the concrete Wrappers!')





import torch
from auto_LiRPA import BoundedTensor
from CTRAIN.util import construct_c

def bound_ibp(model, ptb, data, target, n_classes=10, bound_upper=False, reuse_input=False):
    data = BoundedTensor(data, ptb=ptb)
    if target is not None:
        c = construct_c(data, target, n_classes)
    else:
        c = None
    if reuse_input:
        bound_input = None
    else:
        bound_input = (data,)
    lb, ub = model.compute_bounds(x=bound_input, IBP=True, method="IBP", C=c, bound_upper=bound_upper)
    return lb, ub


#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import time
import os

try:
    from pyscipopt import Model, scip, Eventhdlr, SCIP_RESULT, SCIP_EVENTTYPE, SCIP_PARAMSETTING
    import gurobipy as grb
    # import cvxpy as cp
except ModuleNotFoundError:
    pass


class SCIPVariable:
    def __init__(self,  var):
        self.var = var

    @property
    def lb(self):
        return self.var.getLbLocal()

    @property
    def ub(self):
        return self.var.getUbLocal()

    @property
    def LB(self):
        return self.var.getLbLocal()

    @property
    def UB(self):
        return self.var.getUbLocal()

    @property
    def VarName(self):
        return self.var.name


class SCIPModel(Model):
    def __init__(self,  *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setParam('MIPGap', 1e-4)  # default gurobi setting
        self.setParam('MIPGapAbs', 1e-10)  # default gurobi setting
        _setParam = super(SCIPModel, self).setParam
        # TODO: for Shiqi:
        # 1. Enable RLT Cuts (make its priority very negative?)
        # 2. Disable all primal heuristics (we don't care about the "primal" feasible solution of the MIP; we only care the dual bound).
        # 3. Warm start the solver with adversarial candidates (a good primal solution is important).
        # 4. Change focus of the solver, focusing on finding cuts (no direct option, need to change priorities of several things).
        _setParam('separating/rlt/priority', -10000)
        _setParam('separating/rlt/maxunknownterms', -1)
        _setParam('separating/rlt/maxusedvars', -1)
        _setParam('separating/rlt/maxroundsroot', -1)
        _setParam('separating/rlt/onlyoriginal', False)
        _setParam('separating/rlt/goodscore', 0.5)
        _setParam('separating/rlt/hiddenrlt', True)

    @property
    def status(self, *args, **kwargs):
        # https://github.com/scipopt/PySCIPOpt/blob/769db477ca99260f1a5129c95dd08d9dd61088c1/src/pyscipopt/scip.pyx#L4370
        # https://www.gurobi.com/documentation/9.5/refman/optimization_status_codes.html#sec:StatusCodes
        stat = super().getStatus(*args, **kwargs)
        if stat == "optimal":
            return 2
        elif stat == "timelimit":
            return 9
        elif stat == "infeasible":
            return 3
        elif stat == "unbounded":
            return 5
        elif stat == "userinterrupt":
            return 11
        elif stat == "inforunbd":
            return 4
        elif stat == "nodelimit":
            # return 8
            return 9  # We should return 9 for any limit reached.
        elif stat == "totalnodelimit":
            # return "totalnodelimit"  # no exact match
            return 9
        elif stat == "stallnodelimit":
            # return "stallnodelimit"  # no exact match
            return 9
        elif stat == "gaplimit":
            # return "gaplimit"  # no exact match
            return 9
        elif stat == "memlimit":
            # return "memlimit"  # no exact match
            return 9
        elif stat == "sollimit":
            # return "sollimit"  # no exact match
            return 9
        elif stat == "bestsollimit":
            return 15
        elif stat == "restartlimit":
            # return "restartlimit"
            return 9
        else:
            return "unknown"

    def setObjective(self, expression, sense=None):
        if sense == grb.GRB.MAXIMIZE:
            sense = 'maximize'
        elif sense == grb.GRB.MINIMIZE:
            sense = 'minimize'
        else:
            raise NotImplementedError

        if isinstance(expression, SCIPVariable):
            expression = expression.var
        super().setObjective(expression, sense)

    def addVar(self, lb=0.0, ub=None, obj=0.0, vtype=grb.GRB.CONTINUOUS, name=''):
        vtype_dict = {
            grb.GRB.CONTINUOUS: 'C',
            grb.GRB.BINARY: 'B',
            grb.GRB.INTEGER: 'I',
        }

        # args orders are different by default
        # param vtype: type of the variable: 'C' continuous, 'I' integer, 'B' binary, and 'M' implicit integer
        return super().addVar(lb=lb, ub=ub, obj=obj, vtype=vtype_dict[vtype], name=name)

    def getVarByName(self, name):
        vars = super().getVars()
        for v in vars:
            if name in v.name:
                return SCIPVariable(v)

    def reset(self):
        # TODO: not find exact match
        # super().resetParams()  # This is wrong.
        pass

    def setParam(self, name, value):
        # https://www.scipopt.org/doc-8.0.0/html/PARAMETERS.php
        # https://www.gurobi.com/documentation/9.5/refman/parameters.html#sec:Parameters
        def _cuts_handler(name, val):
            _setParam = super(SCIPModel, self).setParam
            if name == 'Cuts':
                # Global cutting plane switch.
                if val <= 0:
                    # Disable all cuts.
                    print('Disabling all SCIP cutting planes.')
                    _setParam('separating/cgmip/freq', -1)
                    _setParam('separating/clique/freq', -1)
                    _setParam('separating/closecuts/freq', -1)
                    _setParam('separating/flowcover/freq', -1)
                    _setParam('separating/cmir/freq', -1)
                    _setParam('separating/knapsackcover/freq', -1)
                    _setParam('separating/aggregation/freq', -1)
                    _setParam('separating/convexproj/freq', -1)
                    _setParam('separating/disjunctive/freq', -1)
                    _setParam('separating/eccuts/freq', -1)
                    _setParam('separating/gauge/freq', -1)
                    _setParam('separating/gomory/freq', -1)
                    _setParam('separating/strongcg/freq', -1)
                    _setParam('separating/gomorymi/freq', -1)
                    _setParam('separating/impliedbounds/freq', -1)
                    _setParam('separating/interminor/freq', -1)
                    _setParam('separating/intobj/freq', -1)
                    _setParam('separating/mcf/freq', -1)
                    _setParam('separating/minor/freq', -1)
                    _setParam('separating/mixing/freq', -1)
                    _setParam('separating/oddcycle/freq', -1)
                    _setParam('separating/rapidlearning/freq', -1)
                    _setParam('separating/rlt/freq', -1)
                    _setParam('separating/zerohalf/freq', -1)
                    # _setParam('limits/nodes', 1)  # Set this for just solving the root node.
                    _setParam('presolving/maxrestarts', 0)  # Disable restarts.
                else:
                    # Change all cuts parameter to default.
                    _setParam('separating/cgmip/freq', -1)  # Default is never.
                    _setParam('separating/clique/freq', 0)  # Default is root node only.
                    _setParam('separating/closecuts/freq', -1)
                    _setParam('separating/flowcover/freq', 10)
                    _setParam('separating/cmir/freq', 10)
                    _setParam('separating/knapsackcover/freq', 10)
                    _setParam('separating/aggregation/freq', 10)
                    _setParam('separating/convexproj/freq', -1)
                    _setParam('separating/disjunctive/freq', 0)
                    _setParam('separating/eccuts/freq', -1)
                    _setParam('separating/gauge/freq', -1)
                    _setParam('separating/gomory/freq', 10)
                    _setParam('separating/strongcg/freq', 10)
                    _setParam('separating/gomorymi/freq', 10)
                    _setParam('separating/impliedbounds/freq', 10)
                    _setParam('separating/interminor/freq', -1)
                    _setParam('separating/intobj/freq', -1)
                    _setParam('separating/mcf/freq', 0)
                    _setParam('separating/minor/freq', 10)
                    _setParam('separating/mixing/freq', 10)
                    _setParam('separating/oddcycle/freq', -1)
                    _setParam('separating/rapidlearning/freq', 5)
                    _setParam('separating/rlt/freq', 0)
                    _setParam('separating/zerohalf/freq', 10)
            else:
                cuts_dict = {
                        # GurobiOptionName: (SCIPOptionName, SCIPOptionDefault),
                        'GomoryPasses': [('separating/gomory/freq', 10), ('separating/gomorymi/freq', 10)],
                        'RLTCuts': [('separating/rlt/freq', 0)],
                        'FlowCoverCuts': [('separating/flowcover/freq', 10), ('separating/aggregation/freq', 10)],
                        'MIRCuts': [('separating/cmir/freq', 10), ('separating/aggregation/freq', 10)],
                        'ImpliedCuts': [('separating/impliedbounds/freq', 10), ('separating/aggregation/freq', 10)],
                }
                if name in cuts_dict:
                    if value > 0:
                        for setting_item in cuts_dict[name]:
                            scip_opt, default_value = setting_item
                            print(f'Setting {scip_opt} to {default_value}')
                            _setParam(scip_opt, default_value)  # Set to default.
                else:
                    raise NotImplementedError(f"SCIP option {name} not supported.")

        params_dict = {
            'OutputFlag': 'display/verblevel',
            'Threads': 'parallel/maxnthreads',
            'FeasibilityTol': 'numerics/feastol',
            'TimeLimit': 'limits/time',
            'MIPGapAbs': 'limits/absgap',
            'MIPGap': 'limits/gap',
            'Cuts': _cuts_handler,
            'GomoryPasses': _cuts_handler,
            'RLTCuts': _cuts_handler,
            'FlowCoverCuts': _cuts_handler,
            'MIRCuts': _cuts_handler,
            'ImpliedCuts': _cuts_handler,
        }

        if name not in params_dict.keys():
            raise NotImplementedError(f"SCIP option {name} not supported.")

        if name == 'OutputFlag':
            value = 1 if value is False else 4

        if isinstance(params_dict[name], str):
            super().setParam(params_dict[name], value)
        else:
            params_dict[name](name, value)

    def addConstr(self, cons, name=''):
        super().addCons(cons=cons, name=name)

    def getConstrs(self):
        return super().getConss()

    def update(self):
        # Not sure SCIP handles model modifications in a lazy fashion
        # but model.optimize() will update all modifications
        pass

    def getConstrByName(self):
        # TODO
        raise NotImplementedError

    def copy(self):
        new_model = SCIPModel(sourceModel=self)
        return new_model

    @property
    def solcount(self):
        sols = self.getSols()
        return len(sols)

    @property
    def objbound(self):
        return self.getDualbound()

    @property
    def objval(self):
        return self.getPrimalbound()


class EarlyStop(Exception):
    def __init__(self, message):
        print(message)


class GenerateCutsEvent(Eventhdlr):
    def eventinit(self):
        self._call_count = 0
        self.model.catchEvent(SCIP_EVENTTYPE.ROWADDEDSEPA, self)

    def eventexit(self):
        self.model.dropEvent(SCIP_EVENTTYPE.ROWADDEDSEPA, self)

    def eventexec(self, event):
        cuts = self.model.getCuts()
        if len(cuts) > 0:
            self._call_count += 1
            filename = f"cuts_{self._call_count}.txt"
            with open(filename, "w") as f:
                for i, c in enumerate(cuts):
                    f.write(f'row {i:5d}/{len(cuts):5d}: ')
                    f.flush()
                    self.model.printRow(c, f)
            print(f"{len(cuts)} cuts saved to {filename}")


class EarlyStopEvent(Eventhdlr):
    def eventinit(self):
        self.model.catchEvent(SCIP_EVENTTYPE.NODESOLVED, self)

    def eventexit(self):
        self.model.dropEvent(SCIP_EVENTTYPE.NODESOLVED, self)

    def eventexec(self, event):
        # assert event.getType() == SCIP_EVENTTYPE.NODESOLVED
        if self.model.getDualbound() > 1e-5:
            print('Early stop: lower bound > 0!')
            self.model.interruptSolve()
            time.sleep(1)

        elif self.model.getPrimalbound() < -1e-5:
            print('Early stop: upper bound < 0!')
            self.model.interruptSolve()
            time.sleep(1)

        # raise EarlyStop('early stop!')


def event():
    # create solver instance
    s = Model()
    s.hideOutput()
    s.setPresolve(SCIP_PARAMSETTING.OFF)
    eventhdlr = EarlyStopEvent()
    s.includeEventhdlr(eventhdlr, "TestFirstLPevent", "python event handler to catch FIRSTLPEVENT")

    # add some variables
    x = s.addVar("x", obj=1.0)
    y = s.addVar("y", obj=2.0)

    # add some constraint
    s.addCons(x + 2*y >= 5)
    # solve problem
    s.optimize()

    # print solution
    assert round(s.getVal(x)) == 5.0
    assert round(s.getVal(y)) == 0.0


if __name__ == "__main__":
    # event()
    # exit()

    # Test wrapper
    model = SCIPModel("Example")  # model name is optional

    # Test functions
    x = model.addVar(name="x")
    y = model.addVar(name="y")

    model.setParam('OutputFlag', bool(os.environ.get('ALPHA_BETA_CROWN_MIP_DEBUG', False)))
    model.setParam('Threads', 1)
    model.setParam("FeasibilityTol", 2e-5)
    model.setParam('TimeLimit', 10)
    model.setParam('BestBdStop', 1e-5)  # Terminate as long as we find a positive lower bound.
    model.setParam('BestObjStop', 1e-5)  # Terminate as long as we find a adversarial example.
    # model.setParam('Heuristics', 0.5)  # no exact match
    # model.setParam('MIPFocus', 1) # no exact match

    obj = (x-y)
    model.setObjective(obj, grb.GRB.MINIMIZE)
    model.addConstr(x + y == 1)
    model.addConstr(x - y >= 1)
    model.update()

    cons = model.getConstrs()

    model.optimize()  # pass
    print('model.status', model.status)
    model.reset()
    tmp_x = model.getVarByName('x')
    model.objval

    sol = model.getBestSol()
    print("x: {}".format(sol[x]))
    print("y: {}".format(sol[y]))





import json
import os
import torch
import torch.optim as optim
from torch.utils.data import DataLoader
import numpy as np

import shutil

from auto_LiRPA import PerturbationLpNorm
from sklearn.metrics import accuracy_score
from tqdm import tqdm

from CTRAIN.bound import bound_ibp, bound_crown, bound_crown_ibp
from CTRAIN.attacks import pgd_attack
from CTRAIN.complete_verification.abCROWN.util import instances_to_vnnlib, get_abcrown_standard_conf
from CTRAIN.complete_verification.abCROWN.verify import limited_abcrown_eval, abcrown_eval
from CTRAIN.util import export_onnx

def eval_acc(model, test_loader, test_samples=np.inf):
    device = torch.device("cuda" if torch.cuda.is_available() else "mps" if torch.backends.mps.is_available() else "cpu")
    model.eval()
    correct = 0
    total = 0
    with torch.no_grad():
        for data, target in test_loader:
            if total >= test_samples:
                break
            batch_indices = min(len(target), test_samples - total)
            data = data[:batch_indices]
            target = target[:batch_indices]
            
            data, target = data.to(device), target.to(device)
            output = model(data)
            _, predicted = torch.max(output.data, 1)
            total += target.size(0)

            correct += (predicted == target).sum().item()

    test_samples = min(test_samples, total)
    # print(f'Accuracy of the standard model on the first {test_samples} test images: {correct / test_samples:.4f}')
    return correct / test_samples

def eval_ibp(model, eps, data_loader, n_classes=10, test_samples=np.inf, device='cuda'):
    certified = 0
    total_images = 0
    for batch_idx, (data, targets) in tqdm(enumerate(data_loader)):
        if total_images >= test_samples:
            continue
        
        ptb = PerturbationLpNorm(eps=eps, norm=np.inf, x_L=torch.clamp(data - eps, data_loader.min, data_loader.max).to(device), x_U=torch.clamp(data + eps, data_loader.min, data_loader.max).to(device))
        data, targets = data.to(device), targets.to(device)

        lb, ub = bound_ibp(
            model=model,
            ptb=ptb,
            data=data,
            target=targets,
            n_classes=n_classes,
            reuse_input=False
        )
        no_certified = torch.sum((lb > 0).all(dim=1)).item()
        # no_falsified = torch.sum((ub < 0).any(dim=1)).item()
        certified += no_certified
        
        total_images += len(targets)
    
    return certified, total_images

def eval_crown_ibp(model, eps, data_loader, n_classes=10, test_samples=np.inf, device='cuda'):
    certified = 0
    total_images = 0
    for batch_idx, (data, targets) in tqdm(enumerate(data_loader)):
        if total_images >= test_samples:
            continue
        
        ptb = PerturbationLpNorm(eps=eps, norm=np.inf, x_L=torch.clamp(data - eps, data_loader.min, data_loader.max).to(device), x_U=torch.clamp(data + eps, data_loader.min, data_loader.max).to(device))
        data, targets = data.to(device), targets.to(device)
        
        lb, ub = bound_crown_ibp(
            model=model,
            ptb=ptb,
            data=data,
            target=targets,
            n_classes=n_classes,
            reuse_input=False
        )
        no_certified = torch.sum((lb > 0).all(dim=1)).item()
        # no_falsified = torch.sum((ub < 0).any(dim=1)).item()
        certified += no_certified
        
        total_images += len(targets)
    
    return certified, total_images


def eval_crown(model, eps, data_loader, n_classes=10, test_samples=np.inf, device='cuda'):
    # IMPORTANT: Data Loader Batch Size must match Bounding Batch Size when using CROWN for evaluation (not important for IBP)
    crown_data_loader = DataLoader(data_loader.dataset, batch_size=1, shuffle=False)
    crown_data_loader.max, crown_data_loader.min, crown_data_loader.std = data_loader.max, data_loader.min, data_loader.std
    certified = 0
    total_images = 0
    for batch_idx, (data, targets) in tqdm(enumerate(crown_data_loader)):
        if total_images >= test_samples:
            continue
        ptb = PerturbationLpNorm(eps=eps, norm=np.inf, x_L=torch.clamp(data - eps, data_loader.min, data_loader.max).to(device), x_U=torch.clamp(data + eps, data_loader.min, data_loader.max).to(device))
        data, targets = data.to(device), targets.to(device)
        lb, ub = bound_crown(
            model=model,
            ptb=ptb,
            data=data,
            target=targets,
            n_classes=n_classes,
            reuse_input=False
        )
        no_certified = torch.sum((lb > 0).all(dim=1)).item()
        # no_falsified = torch.sum((ub < 0).any(dim=1)).item()
        certified += no_certified
        
        total_images += len(targets)
    
    return certified, total_images

def eval_complete_abcrown(model, eps_std, data_loader, n_classes=10, input_shape=[1, 28, 28], test_samples=np.inf, timeout=1000, no_cores=28, abcrown_batch_size=512, separate_abcrown_process=False, device='cuda'):
    
    no_certified, total_images, certified = eval_adaptive(model=model, eps=eps_std, data_loader=data_loader, n_classes=n_classes, test_samples=test_samples, device=device)
    adv_acc, adv_sample_found = eval_adversarial(model=model, data_loader=data_loader, eps=eps_std, n_classes=n_classes, device=device, test_samples=test_samples, return_adv_indices=True)
    adv_sample_found = torch.tensor(adv_sample_found)
    total_images = 0
        
    batch_size = data_loader.batch_size
    
    std_config = get_abcrown_standard_conf(timeout=timeout, no_cores=no_cores)
    std_config['solver']['batch_size'] = abcrown_batch_size
    
    os.makedirs('/tmp/abCROWN/', exist_ok=True)
    
    export_onnx(
        model=model,
        file_name='/tmp/abCROWN_model.onnx',
        batch_size=1, 
        input_shape=input_shape
    )
    
    for batch_idx, (data, targets) in tqdm(enumerate(data_loader)):
        if total_images >= test_samples:
            break
        
        batch_indices = min(len(targets), test_samples - total_images)
        data = data[:batch_indices]
        targets = targets[:batch_indices]
        total_images += len(targets)
        
        print(f"BATCH {batch_idx}")
        
        clean_pred = torch.argmax(model(data.to(device)), dim=1)
        clean_correct = clean_pred.cpu() == targets
        
        if certified[batch_idx * batch_size:(batch_idx + 1) * batch_size].all():
            continue
        
        os.makedirs(f'/tmp/vnnlib_{batch_idx}/', exist_ok=True)
        
        vnnlib_batch = instances_to_vnnlib(
            indices=[i for i in range(len(targets)) if not certified[batch_idx * batch_size + i] and not adv_sample_found[batch_idx * batch_size + i] and clean_correct[i]],
            data=[(img, target) for img, target in zip(data, targets)],
            vnnlib_path=f'/tmp/vnnlib_{batch_idx}/',
            experiment_name='Experiment',
            eps=eps_std * data_loader.std,
            eps_temp=eps_std,
            data_min=data_loader.min,
            data_max=data_loader.max,
            no_classes=n_classes
        )
        vnnlib_indices = [batch_idx * batch_size + i for i in range(len(targets)) if not certified[batch_idx * batch_size + i] and not adv_sample_found[batch_idx * batch_size + i] and clean_correct[i]]
        print(vnnlib_indices)
        for idx, vnn_instance in zip(vnnlib_indices, vnnlib_batch):
            if separate_abcrown_process:
                running_time, result = limited_abcrown_eval(
                    # work_dir='/tmp/abCROWN',
                    config=std_config,
                    seed=42,
                    instance=vnn_instance,
                    vnnlib_path=f'/tmp/vnnlib_{batch_idx}',
                    model_path=None,
                    model_name=None,
                    model_onnx_path='/tmp/abCROWN_model.onnx',
                    input_shape=[-1] + input_shape[1:4],
                    timeout=timeout,
                    no_cores=no_cores,
                    par_factor=1
                )
            else:
                running_time, result = abcrown_eval(
                    # work_dir='/tmp/abCROWN',
                    config=std_config,
                    seed=42,
                    instance=vnn_instance,
                    vnnlib_path=f'/tmp/vnnlib_{batch_idx}',
                    model_path=None,
                    model_name=None,
                    model_onnx_path='/tmp/abCROWN_model.onnx',
                    input_shape=[-1] + input_shape[1:4],
                    timeout=timeout,
                    no_cores=no_cores,
                    par_factor=1
                )
            print(running_time, result)
            if result == 'unsat':
                no_certified += 1
                certified[idx] = True
            if result == 'sat':
                adv_sample_found[idx] = True
        
        shutil.rmtree(f"/tmp/vnnlib_{batch_idx}/")
    
    if test_samples < np.inf:
        certified = certified[:test_samples]
    no_certified = torch.sum(certified)    
    no_counterexample = torch.sum(adv_sample_found)
            
    certified_acc = (no_certified / test_samples).cpu().item() if torch.is_tensor(certified) else certified / test_samples
    adv_acc = (test_samples  - no_counterexample) / test_samples
    if torch.is_tensor(adv_acc):
        adv_acc = adv_acc.item()
    
    return certified_acc, adv_acc

def eval_adaptive(model, eps, data_loader, n_classes=10, test_samples=np.inf, device='cuda', methods=["IBP", "CROWN_IBP", "CROWN"]):
    assert methods is not None and len(methods) > 1, "Please provide at least one bounding method!"

    certified = torch.tensor([], device=device)
    total_images = 0
    
    crown_data_loader = DataLoader(data_loader.dataset, batch_size=1, shuffle=False)
    crown_data_loader.max, crown_data_loader.min, crown_data_loader.std = data_loader.max, data_loader.min, data_loader.std
    
    for batch_idx, (data, targets) in tqdm(enumerate(data_loader)):
        certified_idx = torch.zeros(len(data), device=device, dtype=torch.bool)
        
        ptb = PerturbationLpNorm(eps=eps, norm=np.inf, x_L=torch.clamp(data - eps, data_loader.min, data_loader.max).to(device), x_U=torch.clamp(data + eps, data_loader.min, data_loader.max).to(device))
        data, targets = data.to(device), targets.to(device)
        if batch_idx * data_loader.batch_size >= test_samples:
            continue
        
        total_images += len(targets)
        
        if "IBP" in methods:
            lb, ub = bound_ibp(
                model=model,
                ptb=ptb,
                data=data,
                target=targets,
                n_classes=n_classes,
                reuse_input=False
            )
            certified_idx[(lb > 0).all(dim=1)] = True
        
        data = data.to('cpu')
        certified_idx = certified_idx.to("cpu")
        ptb = PerturbationLpNorm(eps=eps, norm=np.inf, x_L=torch.clamp(data[~certified_idx] - eps, data_loader.min, data_loader.max).to(device), x_U=torch.clamp(data[~certified_idx] + eps, data_loader.min, data_loader.max).to(device))
        data = data.to(device)
        certified_idx = certified_idx.to(device)
        
        if len(certified_idx) < len(targets) and "CROWN-IBP" in methods:        
            lb, ub = bound_crown_ibp(
                model=model,
                ptb=ptb,
                data=data[~certified_idx],
                target=targets[~certified_idx],
                n_classes=n_classes,
                reuse_input=False
            )
            certified_idx[~certified_idx] = (lb > 0).all(dim=1)
            
        certified = torch.concatenate((certified, certified_idx))
    
    print(f"certified {torch.sum(certified).item()} / {len(certified)} using IBP", flush=True)
    
    for batch_idx, (data, targets) in tqdm(enumerate(crown_data_loader)):
        if batch_idx >= test_samples:
            continue
        if certified[batch_idx] or not ("CROWN" in methods):
            continue
        
        ptb = PerturbationLpNorm(eps=eps, norm=np.inf, x_L=torch.clamp(data - eps, data_loader.min, data_loader.max).to(device), x_U=torch.clamp(data + eps, data_loader.min, data_loader.max).to(device))
        data, targets = data.to(device), targets.to(device)
        
        lb, ub = bound_crown(
            model=model,
            ptb=ptb,
            data=data,
            target=targets,
            n_classes=n_classes,
            reuse_input=False
        )
        instance_certified = (lb > 0).all(dim=1).item()
        certified[batch_idx] = instance_certified
    
    if test_samples < np.inf:
        certified = certified[:test_samples]
    no_certified = torch.sum(certified)            
    total_images = len(certified)
    
    if 'CROWN' in methods:
        print(f"certified {torch.sum(certified).item()} / {len(certified)} after using CROWN", flush=True)
    
    return no_certified, total_images, certified

# TODO: can we maybe spare no_classes?
def eval_certified(model, data_loader, n_classes=10, eps=.3, test_samples=np.inf, method='IBP'):
    device = torch.device("cuda" if torch.cuda.is_available() else "mps" if torch.backends.mps.is_available() else "cpu")
    model.eval()
    certified = 0
    total_images = 0

    if method == "CROWN":
        certified, total_images = eval_crown(model, eps, data_loader, n_classes, test_samples, device)
    elif method == 'IBP':
        certified, total_images = eval_ibp(model, eps, data_loader, n_classes, test_samples, device)
    elif method == 'CROWN-IBP':
        certified, total_images = eval_crown_ibp(model, eps, data_loader, n_classes, test_samples, device)
    elif method == 'ADAPTIVE':
        certified, total_images, cert_results = eval_adaptive(model, eps, data_loader, n_classes, test_samples, device)
    elif isinstance(method, list):
        certified, total_images, cert_results = eval_adaptive(model, eps, data_loader, n_classes, test_samples, device, methods=method)
    elif method == 'COMPLETE':
        # TODO: Infer input shape or pass it, pass timeout and no_cores!
        eval_complete_abcrown(model, eps, data_loader, n_classes, input_shape=(1, 28, 28), test_samples=test_samples, timeout=1000, no_cores=28, device=device)
    elif isinstance(method, (list, tuple, np.ndarray)):
        certified, total_images, cert_results = eval_adaptive(model, eps, data_loader, n_classes, test_samples, device, methods=method)
    else:
        assert False, "UNKNOWN BOUNDING METHOD!"
    
    # print(f"Certified accuracy on test examples for eps {eps}: {certified / total_images}")
    return (certified / total_images).cpu().item() if torch.is_tensor(certified) else certified / total_images

# TODO: Make PGD params adjustable
def eval_adversarial(model, data_loader, eps, n_classes=10, device='cuda', test_samples=np.inf, return_adv_indices=False):
    model.eval()
    adv_preds = np.array([])
    labels = np.array([])
    data_min = data_loader.min.to(device)
    data_max = data_loader.max.to(device)
    
    for batch_idx, (data, targets) in tqdm(enumerate(data_loader)):
        if len(labels) >= test_samples:
            break
        
        batch_indices = min(len(targets), test_samples - len(labels))
        data = data[:batch_indices]
        targets = targets[:batch_indices]
        
        data, targets = data.to(device), targets.to(device)
        eps = eps.to(device)

        x_test_adv = pgd_attack(
            model=model,
            data=data,
            target=targets,
            x_L=torch.clamp(data - eps, data_min, data_max).to(device), 
            x_U=torch.clamp(data + eps, data_min, data_max).to(device),
            restarts=5,
            step_size=.1,
            n_steps=40,
            early_stopping=False,
            device=device
        )
        
        adv_predictions_batch = model(x_test_adv)
        adv_predictions_batch = torch.argmax(adv_predictions_batch, dim=1).cpu().numpy()
        if len(labels) + len(targets) > test_samples:
            too_many_samples_no = (len(labels) + len(targets)) % test_samples
            adv_predictions_batch = adv_predictions_batch[:-too_many_samples_no]
            targets = targets[:-too_many_samples_no]

        adv_preds = np.append(adv_preds, adv_predictions_batch)
        labels = np.append(labels, targets.cpu())

    test_samples = min(test_samples, len(labels))

    adv_accuracy = accuracy_score(labels, adv_preds)
    
    if return_adv_indices:
        adv_sample_found = labels != adv_preds
        return adv_accuracy, adv_sample_found
    # print(f"Accuracy on adversarial test examples for eps {eps}: {adv_accuracy}")
    return adv_accuracy

def eval_model(model, data_loader, n_classes=10, eps=.3, test_samples=np.inf, method='ADAPTIVE', device='cuda'):
    std_acc = eval_acc(model, test_loader=data_loader, test_samples=test_samples)
    cert_acc = eval_certified(model=model, data_loader=data_loader, n_classes=n_classes, eps=eps, test_samples=test_samples, method=method)
    adv_acc = eval_adversarial(model=model, data_loader=data_loader, n_classes=n_classes, eps=eps, device=device, test_samples=test_samples)
    
    return std_acc, cert_acc, adv_acc
    
def eval_epoch(model, data_loader, eps, n_classes, device='cuda', test_samples=1000, verification_method="IBP", results_path="./results"):
    os.makedirs(results_path, exist_ok=True)
    model.eval()
    std_acc = eval_acc(model, test_loader=data_loader, test_samples=test_samples)
    if (eps == 0.).all():
        cert_acc = adv_acc = std_acc
    else:
        with torch.no_grad():
            cert_acc = eval_certified(model=model, data_loader=data_loader, n_classes=n_classes, eps=eps, test_samples=test_samples, method=verification_method)
            adv_acc = eval_adversarial(model=model, data_loader=data_loader, n_classes=n_classes, eps=eps, device=device, test_samples=test_samples)
    
    with open(f"{results_path}/stats.json", "w") as f:
        json.dump(
            {"acc": std_acc, "cert_acc": cert_acc, "adv_acc": adv_acc}, f
        )
    model.train()

    return std_acc, cert_acc, adv_acc 
    



from pathlib import Path
from setuptools import setup, find_packages
from setuptools.command.install import install
import subprocess
import os

def run_command(command, cwd=None):
    result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, cwd=cwd, env=os.environ)
    print(f"Running command: {' '.join(command)}")
    print(f"Return code: {result.returncode}")
    print(f"stdout: {result.stdout}")
    print(f"stderr: {result.stderr}")
    result.check_returncode()

def install_abcrown(installation_path):
    run_command(["touch", "__init__.py"], cwd=installation_path)
    run_command(["touch", "__init__.py"], cwd=f"{installation_path}/complete_verifier/input_split")

class CustomInstallCommand(install):
    def run(self):
        run_command(["pip", "install", "--no-deps", "git+https://github.com/KaidiXu/onnx2pytorch@8447c42c3192dad383e5598edc74dddac5706ee2"])
        run_command(["pip", "install", "--no-deps", "git+https://github.com/Verified-Intelligence/auto_LiRPA.git@cf0169ce6bfb4fddd82cfff5c259c162a23ad03c"])
        install.run(self)
        install_abcrown("./CTRAIN/verification_systems/abCROWN")

setup(
    name="CTRAIN",
    version="0.1.0",
    packages=find_packages(),
    install_requires=[
        "torch>=2,<2.4.0",
        "torchvision>=0.12.0,<0.19.0",
        "numpy>=1.20,<2",
        "tqdm>=4.64",
        "ConfigSpace==0.7.1",
        "smac>=2",
        "pandas>=2",
        "matplotlib>=3",
        "seaborn",
        "onnxruntime>=1",
        "onnxoptimizer",
        "skl2onnx"
    ],
    cmdclass={
        "install": CustomInstallCommand,
    },
    package_data={
        "CTRAIN": ["verification_systems/abCROWN/*"],
        "": ["onnx2pytorch/*", "auto_LiRPA/*"]
    },
    include_package_data=True,
)




import json
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def plot_accs(eps_list, cert_accs, adv_accs):
    x = np.arange(len(eps_list))  # the label locations
    width = 0.4  # the width of the bars
    multiplier = 0

    fig, ax = plt.subplots(constrained_layout=True)
    
    data = {
        "Certified Accuracy": cert_accs,
        "Adversarial Accuracy": adv_accs
    }

    for attribute, measurement in data.items():
        offset = width * multiplier
        rects = ax.bar(x + offset, measurement, width, label=attribute)
        
        labels = [f"{measure:.2f}" for measure in measurement]
        ax.bar_label(rects, labels=labels, padding=3)
        multiplier += 1

    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel('Accuracy')
    ax.set_title('Adv. and Certified Robustness')
    ax.set_xticks(x + width, eps_list)
    ax.legend(loc='upper right', ncols=1,)
    # ax.set_ylim(0, 1)
    
    fig.tight_layout(pad=2)
    
    plt.savefig("./bar_plot_accs.png")


def plot_results_boxplot(stats_path, boxplot_path):
    with open(stats_path, 'r') as f:
        stats = json.load(f)
    
    df = pd.DataFrame.from_dict({(i,j): stats[i][j] 
                           for i in stats.keys() 
                           for j in stats[i].keys()},
                       orient='index')
    df['network_eps'] = df.index.get_level_values(0)

    df_melted = df.melt(id_vars='network_eps', var_name='Evaluation Method', value_name='Accuracy')
    plt.figure(figsize=(20, 10))
    sns.boxplot(x='network_eps', y='Accuracy', data=df_melted, hue='Evaluation Method')
    plt.title("Distribution of Accuracies over 10 Training Runs")
    plt.savefig(boxplot_path)
    
    
if __name__ == "__main__":
    plot_results_boxplot('./results/gowal/mnist/stats.json', 'results/gowal/mnist/stats_boxplot.png')


import torch

from CTRAIN.bound.taps import GradExpander
from CTRAIN.bound import bound_taps
from CTRAIN.train.certified.losses import get_loss_from_bounds


def get_taps_loss(original_model, hardened_model, bounded_blocks, criterion, data, target, n_classes, ptb, device='cuda', pgd_steps=8, pgd_restarts=1, 
                  pgd_step_size=None, pgd_decay_factor=.2, pgd_decay_checkpoints=(5,7), gradient_link_thresh=.5,
                  gradient_link_tolerance=1e-05, gradient_expansion_alpha=5, propagation="IBP", sabr_args=None, return_bounds=False, return_stats=False):
    assert len(bounded_blocks) == 2, "Split not supported!"
    
    taps_bound, ibp_bound = bound_taps(
        original_model=original_model,
        hardened_model=hardened_model,
        bounded_blocks=bounded_blocks,
        data=data,
        target=target,
        n_classes=n_classes,
        ptb=ptb,
        device=device,
        pgd_steps=pgd_steps,
        pgd_restarts=pgd_restarts,
        pgd_step_size=pgd_step_size, 
        pgd_decay_factor=pgd_decay_factor,
        pgd_decay_checkpoints=pgd_decay_checkpoints,
        gradient_link_thresh=gradient_link_thresh,
        gradient_link_tolerance=gradient_link_tolerance,
        propagation=propagation,
        sabr_args=sabr_args
    )
        
    taps_loss = get_loss_from_bounds(taps_bound, criterion)
    ibp_loss = get_loss_from_bounds(ibp_bound, criterion)
    
    loss = GradExpander.apply(taps_loss, gradient_expansion_alpha) * ibp_loss
    
    return_tuple = (loss,)
    
    if return_bounds:
        return_tuple = return_tuple + (taps_bound, None)
    if return_stats:
        robust_err = torch.sum((taps_bound < 0).any(dim=1)).item() / data.size(0)
        return_tuple = return_tuple + (robust_err,)

    return return_tuple



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
from collections import defaultdict
import torch
import numpy as np
from heuristics.babsr import BabsrBranching
from utils import get_reduce_op


class FsbBranching(BabsrBranching):

    @torch.no_grad()
    def get_branching_decisions(
            self, domains, split_depth, branching_candidates=5,
            branching_reduceop='min', use_beta=False, prioritize_alphas='none',
            **kwargs):

        lower_bounds, upper_bounds = domains['lower_bounds'], domains['upper_bounds']
        orig_mask, lAs, cs = domains['mask'], domains['lAs'], domains['cs']
        history = domains['history']
        alphas, betas = domains['alphas'], domains['betas']
        rhs = domains['thresholds']

        batch = len(next(iter(orig_mask.values())))
        # Mask is 1 for unstable neurons. Otherwise it's 0.
        mask = orig_mask
        reduce_op = get_reduce_op(branching_reduceop)
        # In case number of unstable neurons less than topk
        topk = min(branching_candidates,
                   int(sum([item.sum() for item in mask.values()]).item()))
        number_bounds = 1 if cs is None else cs.shape[1]
        score, intercept_tb = self.babsr_score(
            lower_bounds, upper_bounds, lAs, mask, reduce_op,
            number_bounds, prioritize_alphas)

        final_decision = [[] for _ in range(batch)]
        decision_tmp = {}
        tmp_ret = {}
        score_from_layer_idx = 1 if len(score) > 1 else 0
        skip_layers = list(range(score_from_layer_idx))

        # real batch = batch * 2, since we have two kinds of scores
        lbs = {k: torch.concat([v, v]) for k, v in lower_bounds.items()}
        ubs = {k: torch.concat([v, v]) for k, v in upper_bounds.items()}
        # per neuron alpha.
        sps = defaultdict(dict)
        for k, vv in alphas.items():
            sps[k] = {}
            for kk, v in vv.items():
                sps[k][kk] = torch.cat([v, v], dim=2)
        if use_beta:
            bs = [torch.cat([i, i]) for i in betas]
            history += history
        rhs = torch.cat([rhs, rhs])
        if cs is not None:
            cs = torch.cat([cs, cs])
        set_alpha = True  # We only set the alpha once.

        for i in range(score_from_layer_idx, len(score)):
            if ((score[i].max(1).values <= 1e-4).all()
                    and (intercept_tb[i].min(1).values >= -1e-4).all()):
                print(f'{i}th layer has no valid scores')
                skip_layers.append(i)
                continue
            score_idx = torch.topk(score[i], topk)
            score_idx_indices = score_idx.indices.cpu()
            itb_idx = torch.topk(intercept_tb[i], topk, largest=False)
            itb_idx_indices = itb_idx.indices.cpu()
            k_ret = torch.empty(size=(topk, batch * 2), device=score[i].device)
            k_decision = []
            for k in range(topk):
                decision_index = score_idx_indices[:, k]
                # add decision_index with layer's idx
                decision_max_ = [[i, j.item()] for j in decision_index]
                decision_index = itb_idx_indices[:, k]
                decision_min_ = [[i, j.item()] for j in decision_index]
                k_decision.append(decision_max_ + decision_min_)
                # only save the best lower bounds of the two splits
                args_update_bounds = {
                    'lower_bounds': lbs, 'upper_bounds': ubs,
                    'alphas': sps if set_alpha else {}, 'cs': cs
                }
                split = {'decision': k_decision[-1]}
                self.net.build_history_and_set_bounds(args_update_bounds, split)
                if use_beta:
                    args_update_bounds.update({'betas': bs, 'history': history})
                    k_ret_lbs = self.net.update_bounds(
                        args_update_bounds,
                        fix_interm_bounds=True, shortcut=True, beta_bias=False)
                else:
                    k_ret_lbs = self.net.update_bounds(
                        args_update_bounds, beta=False,
                        fix_interm_bounds=True, shortcut=True, beta_bias=False)
                # consider the max improvement among multi bounds in one C matrix
                k_ret_lbs = (k_ret_lbs - torch.cat([rhs, rhs])).max(-1).values
                # No need to set alpha next time; we do not optimize the alphas.
                set_alpha = False
                # build mask indicates invalid scores (stable neurons), batch wise, 1: invalid
                mask_score = (score_idx.values[:, k] <= 1e-4).float()
                mask_itb = (itb_idx.values[:, k] >= -1e-4).float()
                # make the invalid lower bounds worse than normal lower bounds by minus 999999
                # we only consider the best lower bound across two splits by using min(0)
                k_ret[k] = reduce_op((
                    k_ret_lbs.view(-1) - torch.cat(
                        [mask_score, mask_itb]).repeat(2) * 999999
                    ).reshape(2, -1),
                    dim=0).values
            split_depth = min(split_depth, k_ret.shape[0])
            i_idx = k_ret.topk(split_depth, dim=0)  # compare across topK
            tmp_ret[i] = i_idx.values  # [split_depth, batch*2]
            tmp_indice = i_idx.indices
            decision_tmp[i] = [
                k_decision[tmp_indice[ii // (2 * batch)][
                    ii % (2 * batch)]][ii % (2 * batch)]
                for ii in range(split_depth * (batch * 2))
            ]

        # shape of tmp_ret: [layer, num_split, batch*2]
        if len(tmp_ret):
            stacked_layers = torch.stack([i for i in tmp_ret.values()])  # [layer, split_depth, batch*2]
            max_ret = torch.topk(stacked_layers.view(-1, batch * 2), split_depth,
                                 dim=0)  # compare across layers [split_depth, batch*2]
            # shape: [num_split*batch*2]
            rets, decision_layers = max_ret.values.view(-1).cpu().numpy(), max_ret.indices.view(
                -1).cpu().numpy()  # first batch: score; second batch: intercept_tb.
            decision_layers = decision_layers // split_depth

            # add index number for the skipped layers
            # for _, g in groupby(enumerate(skip_layers), lambda ix: ix[0] - ix[1]):
            #     decision_layers[decision_layers >= list(g)[-1][-1]] += 1
            for s in skip_layers:
                decision_layers[decision_layers >= s] += 1

            for l in range(split_depth):
                for b in range(batch):
                    decision_layer_1, decision_index_1 = decision_tmp[
                        decision_layers[2 * l * batch + b].item()][
                        l * 2 * batch + b]
                    decision_layer_2, decision_index_2 = decision_tmp[
                        decision_layers[2 * l * batch + b + batch].item()
                    ][l * 2 * batch + b + batch]
                    decision_layer_1 = self.net.split_nodes[decision_layer_1].name
                    decision_layer_2 = self.net.split_nodes[decision_layer_2].name
                    len_final_decision = len(final_decision[b])
                    if (max([s[b].max() for s in score]) > 1e-4
                            and min([s[b].min() for s in intercept_tb]) < -1e-4
                            and max(rets[2 * l * batch + b], rets[2 * l * batch + b + batch]) > -10000
                            and (mask[decision_layer_1][b][decision_index_1] != 0
                                 or mask[decision_layer_2][b][decision_index_2] != 0)
                    ):  # make sure this potential split is valid
                        if (rets[2 * l * batch + b] > rets[2 * l * batch + b + batch]
                                and mask[decision_layer_1][b][decision_index_1] != 0):  # score > intercept_tb
                            final_decision[b].append(
                                decision_tmp[decision_layers[2 * l * batch + b].item()][l * 2 * batch + b])
                        elif mask[decision_layer_2][b][decision_index_2] != 0:
                            final_decision[b].append(decision_tmp[decision_layers[2 * l * batch + b + batch].item()][
                                                         l * 2 * batch + b + batch])
                        else:
                            mask_item = {k: m[b] for k, m in mask.items()}
                            for preferred_layer in np.random.choice(len(self.net.split_nodes), len(self.net.split_nodes), replace=False):
                                preferred_layer_ = self.net.split_nodes[preferred_layer].name
                                if len(mask_item[preferred_layer_].nonzero(as_tuple=False)) != 0:
                                    final_decision[b].append(
                                        [preferred_layer, mask_item[preferred_layer_].nonzero(as_tuple=False)[0].item()])
                                    break
                    else:
                        # using a random choice
                        mask_item = {k: m[b] for k, m in mask.items()}
                        for preferred_layer in np.random.choice(len(self.net.split_nodes), len(self.net.split_nodes), replace=False):
                            preferred_layer_ = self.net.split_nodes[preferred_layer].name
                            if len(mask_item[preferred_layer_].nonzero(as_tuple=False)) != 0:
                                final_decision[b].append(
                                    [preferred_layer, mask_item[preferred_layer_].nonzero(as_tuple=False)[0].item()])
                                break
                    if len(final_decision[b]) > len_final_decision:
                        final_decision_ = self.net.split_nodes[final_decision[b][-1][0]].name
                        mask[final_decision_][b][final_decision[b][-1][1]] = 0
        else:
            # all layers are split or has no improvement
            for b in range(split_depth * batch):
                # using a random choice
                mask_item = {k: m[b] for k, m in mask.items()}
                for preferred_layer in reversed(self.net.split_nodes):  # from last layer to first layer
                    if len(mask_item[preferred_layer].nonzero(as_tuple=False)) != 0:
                        final_decision[b].append(
                            [preferred_layer, mask_item[preferred_layer].nonzero(as_tuple=False)[0].item()])
                        final_decision_ = self.net.split_nodes[final_decision[b][-1][0]].name
                        mask[final_decision_][b][final_decision[b][-1][1]] = 0
                        break

        split_depth = min([len(d) for d in final_decision])
        final_decision = [[batch[i] for batch in final_decision] for i in
                          range(split_depth)]  # change the order of final decision to split_depth * batch
        final_decision = sum(final_decision, [])

        return final_decision, None, split_depth # None for points



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""Pool of adversarial examples. We use them to suggest subdomains for searching better adversarial examples."""

import math
import torch
import arguments
from auto_LiRPA import BoundedModule
from sortedcontainers import SortedList


class AdvExample:
    """A single adversarial example."""

    def __init__(self, x, obj, pattern):
        self.x = x
        self.obj = obj  # smaller obj is better, we want it to be negative.
        self.activation_pattern = pattern  # A list, each element is the activation pattern of one layer's neuron.

    def __lt__(self, other):
        return self.obj < other.obj

    def __le__(self, other):
        return self.obj <= other.obj

    def __eq__(self, other):
        return self.obj == other.obj


class AdvExamplePool:
    """Keeps a pool of best adversarial examples so far."""

    """We only keep the best `capacity` adversarial images."""
    def __init__(self, network, unstable_mask, capacity=100, C=None):
        assert isinstance(network, BoundedModule)
        self.net = network
        self.unstable_mask = [m.squeeze(0).int().cpu() for m in unstable_mask]  # Initial unstable neuron masks.
        self.total_unstable = sum([m.sum().item() for m in self.unstable_mask])
        self.capacity = 100
        self.adv_pool = SortedList()
        self.C = C  # Output specification in [1, num_of_label_verify, out_dim]
        self.nlayers = len(self.net.relus)
        self.threshold = arguments.Config["bab"]["attack"]["adv_pool_threshold"]
        if self.threshold is not None:
            print('Threshold of adding examples in adv_pool is set to: {} by config file'.format(self.threshold))

    """Add adversarial examples to the pool. Expect adv_images in (N, C, H, W) format."""
    def add_adv_images(self, adv_images, max_to_add=float("inf")):
        if adv_images.size(0) == 0:
            return
        with torch.no_grad():
            # reset_perturbed_nodes must be False otherwise the .perturbed property will be missing.
            pred = self.net(adv_images, reset_perturbed_nodes=False).cpu()  # Pred has size [batch, out_dim]
        adv_images = adv_images.cpu()
        if self.C is not None:
            # Compute the "margin", < 0 means attack successful.
            pred = pred.matmul(self.C.cpu()[0].transpose(-1, -2)).squeeze(-1)
        else:
            assert pred.size(1) == 1
            # Prediction is just one element.
            pred = pred.squeeze(1)
        print("AdvPool received image with prediction:", pred)
        # Get all activations and transfer them to CPU.
        activations = [None] * self.nlayers
        for layer_i, layer in enumerate(self.net.relus):
            # +1 => active, 0 => inactive
            activations[layer_i] = (layer.inputs[0].forward_value.flatten(1) > 0).int().cpu()
        # Set activations for each example.
        c_replaced = 0
        c_added = 0
        c_rejected = 0
        for adv_i in range(adv_images.size(0)):
            if len(self.adv_pool) >= self.capacity:
                if self.adv_pool[-1].obj > pred[adv_i].item():
                    self.adv_pool.pop(-1)  # Remove the worst adversarial example.
                else:
                    # This adversarial example is worse than the worst one in our list. Skip it.
                    c_rejected += 1
                    continue
            example_activations = [None] * self.nlayers
            for layer_i, layer in enumerate(self.net.relus):
                example_activations[layer_i] = activations[layer_i][adv_i]
            c_replaced, c_added, c_rejected = self.replace_adv_example(AdvExample(adv_images[adv_i], pred[adv_i].item(), example_activations), c_replaced, c_added, c_rejected)
        print('Number of adv examples replaced/added/rejected: {}/{}/{}'.format(c_replaced, c_added, c_rejected))

    def print_pool_status(self):
        print('Current adv_pool statistic: length {}, variance {}, Manhattan distance mean {}, min {}, max {}, std {:.4f}'.format(len(self.adv_pool), *self.get_var()))
        for t in [0.6, 0.8, 1.0]:
            act_length = len(self.get_activation_pattern_from_pool(prob_threshold=t)[0])
            print(f'{t*100}% threshold will select {act_length} neurons.')

    """replace the closest adv examples in adv_pool if the diff < threshold"""
    def replace_adv_example(self, adv_example, c_replaced, c_added, c_rejected):
        if len(self.adv_pool) > arguments.Config["attack"]["pgd_restarts"]:
            current_adv_patterns = torch.stack([torch.cat([ii.flatten() for ii in i.activation_pattern]) for i in self.adv_pool])
            this_pattern = torch.cat([ii.flatten() for ii in adv_example.activation_pattern]).view(1, -1)
            diff = torch.cdist(current_adv_patterns.float(), this_pattern.float(), p=0).flatten()
            min_idx = diff.argmin()
            if self.threshold is None:
                # we set the threshold as the lowest diff when the first time we need to filter
                if diff[min_idx] > 1.:
                    self.threshold = diff[min_idx]
                else:
                    self.threshold = 1.
                print('Threshold of adding examples in adv_pool is automatically set to: {}'.format(self.threshold))
            if diff[min_idx] < self.threshold and adv_example.obj < self.adv_pool[min_idx].obj:
                # print('Replace adv_example.')
                self.adv_pool.pop(min_idx)  # pop the closest adv examples in adv_pool
                self.adv_pool.add(adv_example)
                return c_replaced + 1, c_added, c_rejected
            elif adv_example.obj < self.adv_pool[min_idx].obj or diff[min_idx] >= self.threshold:
                # add to adv_pool either the example has better obj or different pattern
                # print('Add better adv_example without replacement.')
                self.adv_pool.add(adv_example)
                return c_replaced, c_added + 1, c_rejected
            else:
                return c_replaced, c_added, c_rejected + 1
        else:
            self.adv_pool.add(adv_example)
            return c_replaced, c_added + 1, c_rejected

    """get the variance of activation patterns in current adv_pool"""
    def get_var(self):
        if len(self.adv_pool) > 0:
            current_adv_patterns = torch.stack([torch.cat([ii.flatten() for ii in i.activation_pattern]) for i in self.adv_pool[1:]])
            best_adv_pattern = torch.cat([ii.flatten() for ii in self.adv_pool[0].activation_pattern]).view(1, -1)
            variance = current_adv_patterns.float().var(0).mean()  # var across length of adv_pool
            distance = torch.cdist(current_adv_patterns.float(), best_adv_pattern.float(), p=0).flatten()
            return variance, distance.mean(), distance.min(), distance.max(), distance.std()
        else:
            return None

    """
    Given an probability threshold, return the common activation shared by at least threshold% adversarial examples.
    If "find_uncommon" is set to True, we find the least common activations, and the coeffs returned are invalid (only decision matters).
    """
    def get_activation_pattern_from_pool(self, prob_threshold=1.0, find_uncommon=False, n_advs=-1):
        all_layer_decisions = []
        all_layer_coeffs = []
        if n_advs == -1:
            # Use entire pool.
            selected_advs = self.adv_pool
        else:
            # Choose the top-n adversarial examples.
            selected_advs = self.adv_pool[:n_advs]
        # Threshold of number of examples with active/inactive neurons.
        pos_threshold = min(int(math.ceil(prob_threshold * len(selected_advs))), len(selected_advs))
        neg_threshold = len(selected_advs) - pos_threshold
        if pos_threshold == neg_threshold:
            # Cannot be equal - otherwise one neuron will be counted in both category.
            neg_threshold -= 1
        # Collect all activation patterns.
        all_patterns = [[] for i in range(self.nlayers)]
        for adv in selected_advs:
            for layer_i in range(self.nlayers):
                all_patterns[layer_i].append(adv.activation_pattern[layer_i])
        # Concatenate activation patterns across examples.
        for layer_i in range(self.nlayers):
            layer_i_pattern = torch.stack(all_patterns[layer_i], dim=0)
            # Shape is (batch, neurons).
            acc_pattern = layer_i_pattern.sum(dim=0)
            # +1 for most neurons active, -1 for most neurons inactive. 0 means this neuron is not selected due to the threshold or unstable mask.
            remaining_acc_pattern = ((acc_pattern >= pos_threshold).int() - (acc_pattern <= neg_threshold).int()) * self.unstable_mask[layer_i]
            if find_uncommon:
                nnz = ((remaining_acc_pattern == 0) * self.unstable_mask[layer_i]).nonzero(as_tuple=True)[0]
            else:
                nnz = remaining_acc_pattern.nonzero(as_tuple=True)[0]
            # Decision is an array with (layer_i, neuron_ids).
            decisions = torch.stack([torch.empty_like(nnz).fill_(layer_i), nnz], dim=1)
            # Decision is an array with +1/-1. (0 when find_uncommon is True).
            coeffs = remaining_acc_pattern[nnz]
            all_layer_decisions.append(decisions)
            all_layer_coeffs.append(coeffs)

        decisions = torch.cat(all_layer_decisions, dim=0)
        coeffs = torch.cat(all_layer_coeffs, dim=0)

        return decisions, coeffs

    """Return a list of unstable neurons, ranked by how common they are in the pool."""
    def get_ranked_activation_pattern(self, n_activations=-1, blacklist=None, find_uncommon=False, n_advs=-1, random_keep=False, softmax_temperature=10.0):
        if n_advs == -1:
            # Use entire pool.
            selected_advs = self.adv_pool
        else:
            # Choose the top-n adversarial examples.
            selected_advs = self.adv_pool[:n_advs]
        if isinstance(blacklist, list):
            blacklist = torch.tensor(blacklist, device='cpu', dtype=torch.long)
        # Collect all activation patterns.
        all_patterns = [[] for i in range(self.nlayers)]
        all_counts = [None] * self.nlayers
        all_status = [None] * self.nlayers  # Dominating neuron activation status per layer.
        for adv in selected_advs:
            for layer_i in range(self.nlayers):
                all_patterns[layer_i].append(adv.activation_pattern[layer_i])
        total_unstable = 0
        # Concatenate activation patterns across examples and count activations.
        for layer_i in range(self.nlayers):
            layer_i_pattern = torch.stack(all_patterns[layer_i], dim=0)
            # Count number of active neurons.
            counts = layer_i_pattern.sum(dim=0)
            # Total number of adv examples.
            total = layer_i_pattern.size(0)
            # Move the counts to reflect both sides (active/inactive).
            # A larger value means more common activations. Stable neurons will have negative counts.
            all_counts[layer_i] = (counts.float() - total / 2).abs()
            enabled_mask = self.unstable_mask[layer_i].clone()
            # Find out the neuron idx that are masked in this layer.
            if blacklist is not None:
                disabled_idx = blacklist[blacklist[:, 0] == layer_i][:, 1].squeeze()
                enabled_mask[disabled_idx] = 0
            # Make disable neurons' counts negative
            all_counts[layer_i] -= 2 * total * (1 - enabled_mask)
            total_unstable += enabled_mask.sum().item()
            # Store the dominating activation status.
            all_status[layer_i] = counts > (total / 2)
        # Find the boundaries of counts.
        counts_length = torch.cumsum(torch.tensor([c.numel() for c in all_counts]), dim=0).tolist()
        counts_length = torch.tensor([0] + counts_length)
        # Flatten the counts tensor.
        flat_all_counts = torch.cat(all_counts, dim=0)
        flat_all_status = torch.cat(all_status, dim=0)
        # Sort the counts, largest ones at the beginning. (negative means disabled neurons; but they should be removed by [:total_untable]).
        sort_idx = flat_all_counts.argsort(descending=True)[:total_unstable]
        # Find out the original neuron layer/idx.
        sorted_layer_idx = torch.searchsorted(counts_length, sort_idx, right=True) - 1
        # Find out the location at each layer.
        sorted_neuron_idx = sort_idx - counts_length[sorted_layer_idx]
        decisions = torch.stack([sorted_layer_idx, sorted_neuron_idx], dim=1)
        # Find out the activation pattern.
        coeffs = (flat_all_status[sort_idx].float() - 0.5) * 2

        # Now we get decisions and coeffs ranked from most common to most uncommon.
        # If we want to keep the most uncommon ones, just reverse the list.
        if find_uncommon:
            decisions = reversed(decisions)
            coeffs = reversed(coeffs)

        if n_activations != -1:
            if random_keep:
                # Randomly keep n_activations, based on commonness.
                scores = flat_all_counts.sort(descending=True).values[:total_unstable]
                if find_uncommon:
                    # Negate the score, so the most uncommon ones become the first.
                    scores = -reversed(scores)
                # Convert scores to probabilities.
                probs = torch.nn.functional.softmax(scores / softmax_temperature, dim=0)
                # Sample based on the probabilities.
                selected_indices = probs.multinomial(n_activations, replacement=False)
                return decisions[selected_indices], coeffs[selected_indices]

            else:
                # Keep most common ones.
                return decisions[:n_activations], coeffs[:n_activations]
        else:
            return decisions, coeffs

    """
    Given a list of unstable neurons, find the top-1 probability configuration based on activation database.
    `decisions` is a list of neurons in [(layer, neuron_idx), (layer, neuron_idx), ... ].
    """
    def find_most_likely_activation(self, decisions, n_advs=-1):
        # This will get the pattern for all unstable neurons.
        adv_decisions, adv_coeffs = self.get_activation_pattern_from_pool(prob_threshold=0.5, n_advs=n_advs)
        adv_decisions = adv_decisions.tolist()
        adv_coeffs = adv_coeffs.tolist()
        coeffs = []
        for d in decisions:
            assert d in adv_decisions
            coeffs.append(adv_coeffs[adv_decisions.index(d)])
        return coeffs

    """Returns activation patterns as a (decisions, coeffs) format, and exclude neurons in blacklist in decisions format."""
    def get_activation_pattern(self, adv_example, blacklist=None):
        if isinstance(blacklist, list):
            blacklist = torch.tensor(blacklist, device='cpu', dtype=torch.long)
        all_layer_decisions = []
        all_layer_coeffs = []
        for layer_idx, layer_pattern in enumerate(adv_example.activation_pattern):
            # layer_pattern is 0 when inactive, 1 active.
            layer_size = len(layer_pattern)
            # Generate the list of all neurons in this layer.
            all_decisions_this_layer = torch.stack([torch.empty(layer_size, dtype=torch.int32).fill_(layer_idx),
                torch.arange(layer_size, dtype=torch.int32)], dim=1)
            enabled_mask = torch.ones(layer_size, dtype=torch.bool)
            enabled_mask = torch.logical_and(enabled_mask, self.unstable_mask[layer_idx])
            if blacklist is not None:
                # Find out the neuron idx that are masked in this layer.
                disabled_mask = blacklist[blacklist[:, 0] == layer_idx][:, 1].squeeze()
                enabled_mask[disabled_mask] = 0
            all_layer_decisions.append(all_decisions_this_layer[enabled_mask])
            # Converting 0, 1 mask to -1, +1 coeffs.
            all_layer_coeffs.append((layer_pattern[enabled_mask] - 0.5) * 2)

        # Concat all layer decisions and coeffs into 1.
        decisions = torch.cat(all_layer_decisions, dim=0)
        coeffs = torch.cat(all_layer_coeffs, dim=0)

        return decisions, coeffs

    """
    Cross-over activation patterns from different adversarial examples.
    """
    def generate_crossover(self):
        # TODO.
        pass




from concurrent.futures import ProcessPoolExecutor
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
from auto_LiRPA import BoundedModule

from auto_LiRPA.perturbations import PerturbationLpNorm
from CTRAIN.eval import eval_acc, eval_certified, eval_epoch, eval_adversarial
from CTRAIN.bound import bound_ibp
from CTRAIN.train.certified.eps_scheduler import SmoothedScheduler
from CTRAIN.train.certified.losses import get_sabr_loss
from CTRAIN.train.certified.initialisation import ibp_init_shi
from CTRAIN.train.certified.regularisers.shi import get_shi_regulariser
from CTRAIN.util import save_checkpoint
from CTRAIN.train.certified.regularisers import get_l1_reg


# TODO: Add Shi regularisation weight parameter
def sabr_train_model(original_model, hardened_model, train_loader, val_loader=None, num_epochs=None, eps=0.3, eps_std=0.3, eps_schedule=(0, 20, 50), eps_schedule_unit='epoch', eps_scheduler_args=dict(), optimizer=None,
                    subselection_ratio=.4, lr_decay_schedule=(15, 25), lr_decay_factor=10, lr_decay_schedule_unit='epoch', 
                    n_classes=10, gradient_clip=None, l1_regularisation_weight=0.00001, shi_regularisation_weight=1, shi_reg_decay=True,
                    pgd_steps=8, pgd_step_size=.5, pgd_restarts=1, pgd_early_stopping=True, pgd_decay_factor=.1, pgd_decay_checkpoints=(4,7), results_path="./results", device='cuda'):
    

    criterion = nn.CrossEntropyLoss(reduction='none')

    # Important Change to Vanilla IBP: Initialise Weights to normal distribution with sigma_i = sqrt(2*pi)/n_i, for layer i and fan in n_i
    ibp_init_shi(original_model, hardened_model)

    no_batches = 0
    cur_lr = optimizer.param_groups[-1]['lr']

    # Important Change to Vanilla IBP: Schedule Eps smoothly
    eps_scheduler = SmoothedScheduler(
        num_epochs=num_epochs,
        eps=eps,
        mean=train_loader.mean,
        std=train_loader.std,
        eps_schedule_unit=eps_schedule_unit,
        eps_schedule=eps_schedule,
        batches_per_epoch=len(train_loader),
        **eps_scheduler_args
    )

    cur_eps = eps_scheduler.get_cur_eps()
    # Training loop
    for epoch in range(num_epochs):
        
        epoch_adv_err = 0
        epoch_rob_err = 0
        epoch_nat_err = 0
        
        if lr_decay_schedule_unit == 'epoch':
            if epoch + 1 in lr_decay_schedule:
                print("LEARNING RATE DECAYED!")
                cur_lr = cur_lr * lr_decay_factor
                for g in optimizer.param_groups:
                    g['lr'] = cur_lr

        
        print(f"[{epoch + 1}/{num_epochs}]: eps {[channel_eps for channel_eps in cur_eps]}")
        hardened_model.train()
        original_model.train()
        running_loss = 0.0
        
        for batch_idx, (data, target) in enumerate(train_loader):

            cur_eps = eps_scheduler.get_cur_eps().reshape(-1, 1, 1)
            
            ptb = PerturbationLpNorm(eps=cur_eps, norm=np.inf, x_L=torch.clamp(data - cur_eps, train_loader.min, train_loader.max).to(device), x_U=torch.clamp(data + cur_eps, train_loader.min, train_loader.max).to(device))

              
            if lr_decay_schedule_unit == 'batch':
                if no_batches + 1 in lr_decay_schedule:
                    print("LEARNING RATE DECAYED!")
                    cur_lr = cur_lr * lr_decay_factor
                    for g in optimizer.param_groups:
                        g['lr'] = cur_lr
            
            data, target = data.to(device), target.to(device)
            optimizer.zero_grad()
            clean_output = hardened_model(data)
            regular_err = torch.sum(torch.argmax(clean_output, dim=1) != target).item() / data.size(0)
            epoch_nat_err += regular_err
            if eps_scheduler.get_cur_eps(normalise=False) != 0.:
                
                sabr_loss, robust_err, adv_err = get_sabr_loss(
                    hardened_model=hardened_model,
                    original_model=original_model,
                    train_loader=train_loader,
                    data=data,
                    data_min=train_loader.min.to(device),
                    data_max=train_loader.max.to(device),
                    target=target,
                    eps=torch.tensor(cur_eps, device=device),
                    subselection_ratio=subselection_ratio,
                    criterion=criterion,
                    device=device,
                    n_classes=n_classes,
                    pgd_steps=pgd_steps,
                    pgd_step_size=pgd_step_size,
                    pgd_restarts=pgd_restarts,
                    pgd_early_stopping=pgd_early_stopping,
                    pgd_decay_factor=pgd_decay_factor,
                    pgd_decay_checkpoints=pgd_decay_checkpoints,
                    return_stats=True
                )
                
                epoch_adv_err += adv_err
                epoch_rob_err += robust_err

                loss = sabr_loss
            
            else:
                loss = criterion(clean_output, target).mean()

            if (eps_scheduler.get_cur_eps(normalise=False) != eps_scheduler.get_max_eps(normalise=False)):
                # SABR also uses Shi regularisation during warm up/ramp-up
                loss_regularisers = get_shi_regulariser(
                    model=hardened_model,
                    ptb=ptb,
                    data=data,
                    target=target,
                    eps_scheduler=eps_scheduler,
                    n_classes=n_classes,
                    device=device,
                    included_regularisers=['relu', 'tightness'],
                    verbose=False,
                    regularisation_decay=shi_reg_decay
                )
                loss = loss + shi_regularisation_weight * loss_regularisers
            
            if l1_regularisation_weight is not None:
                l1_regularisation = l1_regularisation_weight * get_l1_reg(model=original_model, device=device)
                loss = loss + l1_regularisation
                
            loss.backward()
            if gradient_clip is not None:
                nn.utils.clip_grad_value_(hardened_model.parameters(), clip_value=gradient_clip)
            optimizer.step()
            
            running_loss += loss.item()
            eps_scheduler.batch_step()
            no_batches += 1

        train_acc_nat = (1 - epoch_nat_err / len(train_loader))
        train_acc_adv = (1 - epoch_adv_err / len(train_loader))
        train_acc_cert = (1 - epoch_rob_err / len(train_loader))

        print(f'Epoch [{epoch+1}/{num_epochs}], Train Loss: {running_loss/len(train_loader):.4f}')
        print(f'\t Natural Acc. Train: {train_acc_nat:.4f}')
        print(f'\t Adv. Acc. Train: {train_acc_adv:.4f}')
        print(f'\t Certified Acc. Train: {train_acc_cert:.4f}')

        if results_path is not None:
            save_checkpoint(hardened_model, optimizer, running_loss, epoch + 1, results_path)
        
    return hardened_model

            


#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
"""Optimizing computation graph in onnx file."""
import torch
import onnx
from onnx import numpy_helper as nh
import onnxruntime as ort
import numpy as np


def convert_onnx_to_double(onnx_file):
    """Convert onnx file from single precision to double precision."""
    onnx_model = onnx.load(onnx_file)

    def _convert_data(obj):
        fp64_initializer = nh.to_array(obj).astype(np.float64)
        obj.data_type = onnx.TensorProto.DOUBLE
        obj.raw_data = nh.from_array(fp64_initializer).raw_data

    # Change all initializer to fp64.
    for node in onnx_model.graph.initializer:
        _convert_data(node)
    # Change all inputs to fp64.
    for node in onnx_model.graph.input:
        node.type.tensor_type.elem_type = onnx.TensorProto.DOUBLE
    # Change all outputs to fp64.
    for node in onnx_model.graph.output:
        node.type.tensor_type.elem_type = onnx.TensorProto.DOUBLE
    # Change node attributes to fp64.
    for node in onnx_model.graph.node:
        for attr in node.attribute:
            if hasattr(attr, 't') and hasattr(attr.t, 'raw_data') and attr.t.raw_data:
                _convert_data(attr.t)
    # for node in onnx_model.graph.nodes:
    print('Model converted to double precision.')
    return onnx_model


def fuse_conv_and_bn(conv, bn):
    # init
    fusedconv = torch.nn.Conv2d(
        conv.in_channels,
        conv.out_channels,
        kernel_size=conv.kernel_size,
        stride=conv.stride,
        padding=conv.padding,
        bias=True
    )
    #
    # prepare filters
    w_conv = conv.weight.clone().view(conv.out_channels, -1)
    w_bn = torch.diag(bn.weight.div(torch.sqrt(bn.eps + bn.running_var)))
    fusedconv.weight.copy_(torch.mm(w_bn, w_conv).view(fusedconv.weight.size()))
    #
    # prepare spatial bias
    if conv.bias is not None:
        b_conv = conv.bias
    else:
        b_conv = torch.zeros(conv.weight.size(0))
    b_bn = bn.bias - bn.weight.mul(bn.running_mean).div(torch.sqrt(bn.running_var + bn.eps))
    fusedconv.bias.copy_(torch.matmul(w_bn, b_conv) + b_bn)
    #
    # we're done
    return fusedconv


def create_initializer_tensor(
        name: str,
        tensor_array: np.ndarray,
        data_type: onnx.TensorProto = onnx.TensorProto.FLOAT
) -> onnx.TensorProto:
    # (TensorProto)
    initializer_tensor = onnx.helper.make_tensor(
        name=name,
        data_type=data_type,
        dims=tensor_array.shape,
        vals=tensor_array.flatten().tolist())

    return initializer_tensor


def conv_output_shape(h_w, kernel_size=1, stride=1, pad=0, dilation=1):
    if type(h_w) is not tuple:
        h_w = (h_w, h_w)
    if type(kernel_size) is not tuple:
        kernel_size = (kernel_size, kernel_size)
    if type(stride) is not tuple:
        stride = (stride, stride)
    if type(pad) is not tuple:
        pad = (pad, pad)
    h = (h_w[0] + (2 * pad[0]) - (dilation * (kernel_size[0] - 1)) - 1) // stride[0] + 1
    w = (h_w[1] + (2 * pad[1]) - (dilation * (kernel_size[1] - 1)) - 1) // stride[1] + 1
    return h, w


def strided_conversion(input_feature, kernel, padding, stride):
    output_channel, input_channel, kernel_h, kernel_w = kernel.shape
    input_h, input_w = input_feature.shape[-2:]
    output_h, output_w = conv_output_shape(h_w=(input_h, input_w), kernel_size=(kernel_h, kernel_w), stride=stride,
                                           pad=padding)
    total_locs = output_h * output_w  # total number of output locations.

    converted_matrix = torch.zeros(output_channel, total_locs, input_channel,
                                   (input_h + 2 * padding) * (input_w + 2 * padding)).type(torch.DoubleTensor)
    orig_stride = converted_matrix.stride()

    matrix_strided = torch.as_strided(converted_matrix,
                                      [output_channel, total_locs, output_h, output_w, input_channel, kernel_h,
                                       kernel_w],
                                      [orig_stride[0], orig_stride[1], (input_h + 2 * padding) * stride, stride,
                                       orig_stride[2], input_h + 2 * padding, 1])

    first_indices = torch.arange(total_locs)
    second_indices = torch.div(first_indices, output_h, rounding_mode="trunc")
    third_indices = torch.fmod(first_indices, output_h)
    matrix_strided[:, first_indices, second_indices, third_indices, :, :, :] = kernel.reshape(output_channel, 1,
                                                                                              input_channel, kernel_h,
                                                                                              kernel_w)
    converted_matrix = converted_matrix.view(output_channel * total_locs, input_channel, input_h + 2 * padding,
                                             input_w + 2 * padding)
    if padding > 0:
        converted_matrix = converted_matrix[:, :, padding:-padding, padding:-padding]
    # Output shape is (outC, outH, outW, inC, inH, inW)
    return converted_matrix.reshape(output_channel * output_h * output_w,
                                    converted_matrix.shape[1] * converted_matrix.shape[2] * converted_matrix.shape[3])


def fuse_cgan_gemm_reshape_bn(onnx_model, initializers):
    old_linear_weight = initializers[onnx_model.graph.node[0].input[1]]
    old_linear_bias = initializers[onnx_model.graph.node[0].input[2]]
    old_bn_weight = initializers[onnx_model.graph.node[3].input[1]]
    old_bn_bias = initializers[onnx_model.graph.node[3].input[2]]
    old_bn_mean = initializers[onnx_model.graph.node[3].input[3]]
    old_bn_var = initializers[onnx_model.graph.node[3].input[4]]

    dim_flattened = old_linear_bias.shape[0]
    dim_channel = old_bn_weight.shape[0]
    new_linear_weight = old_linear_weight.reshape(dim_channel, int(dim_flattened/dim_channel), -1).transpose(2, 1, 0) * old_bn_weight / np.sqrt(old_bn_var + onnx_model.graph.node[3].attribute[0].f)
    new_linear_bias = old_linear_bias.reshape(dim_channel, int(dim_flattened/dim_channel)).transpose(1, 0) * old_bn_weight / np.sqrt(old_bn_var + onnx_model.graph.node[3].attribute[0].f) + old_bn_bias - old_bn_weight * old_bn_mean / np.sqrt(old_bn_var + onnx_model.graph.node[3].attribute[0].f)
    new_linear_weight = new_linear_weight.transpose(0, 2, 1).reshape(-1, dim_flattened).transpose(1, 0)
    new_linear_bias = new_linear_bias.transpose(1, 0).reshape(dim_flattened)

    return new_linear_weight, new_linear_bias


def fuse_cgan_bn_reshape_gemm(onnx_model, initializers):
    old_linear_weight = initializers[onnx_model.graph.node[-1].input[1]]
    old_linear_bias = initializers[onnx_model.graph.node[-1].input[2]]
    old_bn_weight = initializers[onnx_model.graph.node[-4].input[1]]
    old_bn_bias = initializers[onnx_model.graph.node[-4].input[2]]
    old_bn_mean = initializers[onnx_model.graph.node[-4].input[3]]
    old_bn_var = initializers[onnx_model.graph.node[-4].input[4]]

    dim_flattened = old_linear_weight.shape[1]
    dim_channel = old_bn_weight.shape[0]
    old_bn_weight_flatten = old_bn_weight[:, None].repeat(4, axis=-1).reshape(dim_flattened)
    old_bn_bias_flatten = old_bn_bias[:, None].repeat(4, axis=-1).reshape(dim_flattened)
    old_bn_mean_flatten = old_bn_mean[:, None].repeat(4, axis=-1).reshape(dim_flattened)
    old_bn_var_flatten = old_bn_var[:, None].repeat(4, axis=-1).reshape(dim_flattened)

    new_linear_weight = old_linear_weight * old_bn_weight_flatten / np.sqrt(old_bn_var_flatten + onnx_model.graph.node[-4].attribute[0].f)
    new_linear_bias = np.matmul(old_linear_weight, (old_bn_bias_flatten - old_bn_weight_flatten * old_bn_mean_flatten / np.sqrt(old_bn_var_flatten + onnx_model.graph.node[-4].attribute[0].f))) + old_linear_bias

    return new_linear_weight, new_linear_bias


def compress_onnx(onnx_model, old_path, save_path, onnx_optimization_flags, debug=False):
    assert onnx_optimization_flags != 'none'
    if debug:
        place_holder = []
        for x in onnx_model.graph.input[0].type.tensor_type.shape.dim:
            place_holder.append(x.dim_value)

        place_holder[0] = 1
        place_holder = tuple(place_holder)

    cur_W, cur_b = None, None
    source = onnx_model.graph.input[0].name

    initializers = {}
    for onnx_module in onnx_model.graph.initializer:
        initializers[onnx_module.name] = nh.to_array(onnx_module)

    new_initializers = []
    new_nodes = []

    cnt, convcnt = 0, 0

    if "merge_gemm" in onnx_optimization_flags:  # bias only. need to generalize this later.
        W1 = initializers[onnx_model.graph.node[0].input[1]]
        b1 = initializers[onnx_model.graph.node[0].input[2]]

        W2 = initializers[onnx_model.graph.node[3].input[1]]
        b2 = initializers[onnx_model.graph.node[3].input[2]]

        input_feature = torch.randn((1, 3, 32, 32)).type(torch.DoubleTensor)
        pads = 1
        strides = 1

        nW = strided_conversion(input_feature, torch.from_numpy(W2).type(torch.DoubleTensor), pads,
                                strides).detach().numpy().astype(np.float32)
        nb = np.repeat(b2, 32 * 32)

        W3 = np.matmul(nW, W1)
        b3 = np.matmul(nW, b1) + nb

        cur_linear_W = create_initializer_tensor(
            name='linear_W',
            tensor_array=W3,
            data_type=onnx.TensorProto.FLOAT)

        new_initializers.append(cur_linear_W)

        cur_linear_b = create_initializer_tensor(
            name='linear_b',
            tensor_array=b3,
            data_type=onnx.TensorProto.FLOAT)

        new_initializers.append(cur_linear_b)

        cur_node = onnx.helper.make_node(
            name='linear_MatMul',
            op_type='Gemm',
            inputs=[onnx_model.graph.node[0].input[0], 'linear_W', 'linear_b'],
            outputs=[onnx_model.graph.node[3].output[0]],
            alpha=1.0,
            beta=1.0,
            transB=1
        )

        new_nodes.append(cur_node)

        re_node = onnx.helper.make_node(
            name='Relu_4',
            op_type='Relu',
            inputs=[onnx_model.graph.node[3].output[0]],
            outputs=['after_relu']
        )

        new_nodes.append(re_node)

        values = np.array([1, 16, 32, 32])
        cnode = onnx.helper.make_node(
            'Constant',
            inputs=[],
            outputs=['values'],
            value=onnx.helper.make_tensor(
                name='const_tensor',
                data_type=onnx.TensorProto.INT64,
                dims=values.shape,
                vals=values.flatten().astype(int),
            ),
        )

        new_nodes.append(cnode)

        rnode = onnx.helper.make_node(
            'Reshape',
            inputs=['after_relu', 'values'],
            outputs=[onnx_model.graph.node[4].output[0]],
        )

        new_nodes.append(rnode)

        started_node = 5

    elif "merge_gemm_reshape_bn" in onnx_optimization_flags:
        # TODO (assigned to Zhuowen): generalize the code to allow the merge of gemm + BN for any general onnx.
        merge_case = (onnx_model.graph.node[0].op_type == "Gemm" and 3 < len(onnx_model.graph.node) and
                onnx_model.graph.node[1].op_type == "Constant" and
                onnx_model.graph.node[2].op_type == "Reshape" and onnx_model.graph.node[3].op_type == "BatchNormalization")
        if merge_case:
            new_linear_weight, new_linear_bias = fuse_cgan_gemm_reshape_bn(onnx_model, initializers)

            linear_W_initializer = create_initializer_tensor(
                name='linear_W',
                tensor_array=new_linear_weight,
                data_type=onnx.TensorProto.FLOAT)
            new_initializers.append(linear_W_initializer)
            linear_b_initializer = create_initializer_tensor(
                name='linear_b',
                tensor_array=new_linear_bias,
                data_type=onnx.TensorProto.FLOAT)
            new_initializers.append(linear_b_initializer)

            cur_node = onnx.helper.make_node(
                op_type='Gemm',
                inputs=[onnx_model.graph.node[0].input[0], 'linear_W', 'linear_b'],
                outputs=['after_linear'],
                alpha=1.0,
                beta=1.0,
                transB=1
            )
            new_nodes.append(cur_node)

            values = nh.to_array(onnx_model.graph.node[1].attribute[0].t)
            cnode = onnx.helper.make_node(
                'Constant',
                inputs=[],
                outputs=['values'],
                value=onnx.helper.make_tensor(
                    name='const_tensor',
                    data_type=onnx.TensorProto.INT64,
                    dims=values.shape,
                    vals=values.flatten().astype(int),
                ),
            )
            new_nodes.append(cnode)
            rnode = onnx.helper.make_node(
                'Reshape',
                inputs=['after_linear', 'values'],
                outputs=[onnx_model.graph.node[3].output[0]],
            )
            new_nodes.append(rnode)

            started_node = 4
        else:
            started_node = 0
    else:
        started_node = 0

    skipped = {}
    const_var = {}
    added_const = {}

    def trace(node_name):
        if node_name in skipped:
            return trace(skipped[node_name])
        return node_name

    if "remove_matmul_inplace" in onnx_optimization_flags:
        to_transpose_const, temp_dict = {}, {}
        for i, node in enumerate(onnx_model.graph.node[started_node:]):
            if node.op_type == "Constant":
                temp_dict[node.output[0]] = i
            matmul_case1 = (node.op_type == "MatMul" and i - 1 >= 0 and i + 1 < len(onnx_model.graph.node) and
                    onnx_model.graph.node[i - 1].op_type == "Transpose" and onnx_model.graph.node[i + 1].op_type == "Transpose" and
                    onnx_model.graph.node[i - 1].output[0] == node.input[1])
            matmul_case2 = (node.op_type == "MatMul" and i - 2 >= 0 and i + 1 < len(onnx_model.graph.node) and
                    onnx_model.graph.node[i - 2].op_type == "Transpose" and onnx_model.graph.node[i + 1].op_type == "Transpose" and
                    onnx_model.graph.node[i - 2].output[0] == node.input[1])
            if matmul_case1 or matmul_case2:
                to_transpose_const[i] = temp_dict[node.input[0]]
    elif "merge_vit" in onnx_optimization_flags:
        to_merge_linears, search_list, to_merge_others = [], [], []
        for i, node in enumerate(onnx_model.graph.node[started_node:]):
            if (node.op_type == "Transpose" and onnx_model.graph.node[i - 1].op_type == "BatchNormalization" and
                onnx_model.graph.node[i - 2].op_type == "Transpose"):
                search_list.append(node.output[0])
            elif node.op_type == "MatMul" and i + 1 < len(onnx_model.graph.node) and onnx_model.graph.node[i + 1].op_type == "Add":
                if node.input[0] in search_list:
                    to_merge_linears.append(i)
            elif node.op_type == "Shape" and node.input[0] in search_list:
                to_merge_others.append(i)

    for i, node in enumerate(onnx_model.graph.node[started_node:]):
        if "remove_ineffective_layers" in onnx_optimization_flags:
            if node.op_type == "Constant":
                val = nh.to_array(node.attribute[0].t)
                mn, mx = np.min(val.reshape(-1)), np.max(val.reshape(-1))
                if (mx - mn <= 1e-9).all():
                    const_var[node.output[0]] = (val, node)
                else:
                    added_const[node.output[0]] = True
                    new_nodes.append(node)
                continue
            elif node.op_type == "Sub":
                if node.input[1] in const_var:
                    if (np.abs(const_var[node.input[1]][0]) <= 1e-9).all():
                        skipped[node.output[0]] = trace(node.input[0])
                        print('Remove ineffective operation: sub(0).')
                        continue
                if node.input[1] not in added_const:
                    added_const[node.input[1]] = True
                    new_nodes.append(const_var[node.input[1]][1])
                # node.input[0] = trace(node.input[0])
                new_nodes.append(node)
                continue
            elif node.op_type == "Div":
                if node.input[1] in const_var:
                    if np.abs(const_var[node.input[1]][0] - 1) <= 1e-9:
                        skipped[node.output[0]] = trace(node.input[0])
                        print('Remove ineffective operation: div(0).')
                        continue
                if node.input[1] not in added_const:
                    added_const[node.input[1]] = True
                    new_nodes.append(const_var[node.input[1]][1])
                node.input[0] = trace(node.input[0])
                new_nodes.append(node)
                continue
        if "merge_bn" in onnx_optimization_flags:
            if node.op_type == "Conv":
                if i + 1 < len(onnx_model.graph.node) and onnx_model.graph.node[i + 1].op_type == "BatchNormalization":
                    bn_node = onnx_model.graph.node[i + 1]
                    w_conv = initializers[node.input[1]]
                    bn_weight = initializers[bn_node.input[1]]
                    bn_bias = initializers[bn_node.input[2]]
                    bn_mean = initializers[bn_node.input[3]]
                    bn_var = initializers[bn_node.input[4]]
                    w_shape = w_conv.shape
                    bn_eps = bn_node.attribute[0].f
                    w_bn = np.diag(bn_weight / (np.sqrt(bn_eps + bn_var)))

                    w_conv = w_conv.reshape(w_shape[0], -1)
                    if len(node.input) == 3:
                        b_conv = node.input[2]
                    else:
                        b_conv = np.zeros(w_shape[0])

                    b_bn = bn_bias - bn_weight * bn_mean / np.sqrt(bn_var + bn_eps)

                    W = np.matmul(w_bn, w_conv).reshape(*w_shape)
                    b = np.matmul(w_bn, b_conv) + b_bn

                    triggered = True

                    conv_W_initializer = create_initializer_tensor(
                        name=f"conv{convcnt}_W",
                        tensor_array=W,
                        data_type=onnx.TensorProto.FLOAT)

                    conv_b_initializer = create_initializer_tensor(
                        name=f"conv{convcnt}_b",
                        tensor_array=b,
                        data_type=onnx.TensorProto.FLOAT
                    )
                    conv_node = onnx.helper.make_node(
                        name=f"conv{convcnt}",
                        op_type="Conv",
                        inputs=[node.input[0], f"conv{convcnt}_W", f"conv{convcnt}_b"],
                        outputs=[bn_node.output[0]],
                        dilations=tuple(node.attribute[0].ints),
                        kernel_shape=tuple(node.attribute[2].ints),
                        pads=tuple(node.attribute[3].ints),
                        strides=tuple(node.attribute[4].ints)
                    )
                    new_nodes.append(conv_node)
                    new_initializers.append(conv_W_initializer)
                    new_initializers.append(conv_b_initializer)
                    convcnt += 1
                else:
                    new_nodes.append(node)
                    source = node.output[0]
                    for old_init in node.input:
                        if old_init in initializers:
                            cur_init = create_initializer_tensor(
                                name=old_init,
                                tensor_array=initializers[old_init],
                                data_type=onnx.TensorProto.FLOAT)
                            new_initializers.append(cur_init)
                    triggered = False
            elif node.op_type == "BatchNormalization":
                if triggered == True:
                    triggered = False
                    continue
                new_nodes.append(node)
                source = node.output[0]
                for old_init in node.input:
                    if old_init in initializers:
                        cur_init = create_initializer_tensor(
                            name=old_init,
                            tensor_array=initializers[old_init],
                            data_type=onnx.TensorProto.FLOAT)
                        new_initializers.append(cur_init)
            else:
                new_nodes.append(node)
                source = node.output[0]
                for old_init in node.input:
                    if old_init in initializers:
                        cur_init = create_initializer_tensor(
                            name=old_init,
                            tensor_array=initializers[old_init],
                            data_type=onnx.TensorProto.FLOAT)
                        new_initializers.append(cur_init)
        elif "merge_linear" in onnx_optimization_flags:
            if node.op_type == "MatMul" or node.op_type == "Add":
                if node.input[-1] in initializers:
                    layer_mat = initializers[node.input[-1]]
                else:
                    new_nodes.append(node)
                    source = node.output[0]
                    for old_init in node.input:
                        if old_init in initializers:
                            cur_init = create_initializer_tensor(
                                name=old_init,
                                tensor_array=initializers[old_init],
                                data_type=onnx.TensorProto.FLOAT)
                            new_initializers.append(cur_init)
                    continue
                if node.op_type == "MatMul":
                    if cur_W is None:
                        cur_W = layer_mat
                    else:
                        cur_W = np.matmul(cur_W, layer_mat)
                        if cur_b is not None: cur_b = np.matmul(cur_b, layer_mat)
                elif node.op_type == "Add":
                    if cur_b is None:
                        cur_b = layer_mat
                    else:
                        layer_b = layer_mat
                        cur_b += layer_b
                target = node.output[0]

                if (i == len(onnx_model.graph.node) - 1 or (
                        onnx_model.graph.node[i + 1].op_type != "MatMul" and
                        onnx_model.graph.node[i + 1].op_type != "Add")):
                    if cur_W is not None:
                        cur_linear_W = create_initializer_tensor(
                            name=f'linear{cnt}_W',
                            tensor_array=cur_W,
                            data_type=onnx.TensorProto.FLOAT)
                        output_node = target
                        if cur_b is not None: output_node = f'linear{cnt}_intermediate'
                        cur_node = onnx.helper.make_node(
                            name=f'linear{cnt}_MatMul',
                            op_type='MatMul',
                            inputs=[source, f'linear{cnt}_W'],
                            outputs=[output_node]
                        )
                        new_initializers.append(cur_linear_W)
                        new_nodes.append(cur_node)
                    if cur_b is not None:
                        cur_linear_b = create_initializer_tensor(
                            name=f'linear{cnt}_b',
                            tensor_array=cur_b,
                            data_type=onnx.TensorProto.FLOAT)

                        input_node = source
                        if cur_W is not None: input_node = f'linear{cnt}_intermediate'

                        cur_node = onnx.helper.make_node(
                            name=f'linear{cnt}_Add',
                            op_type='Add',
                            inputs=[input_node, f'linear{cnt}_b'],
                            outputs=[target]
                        )
                        new_initializers.append(cur_linear_b)
                        new_nodes.append(cur_node)
                    if cur_W is not None or cur_b is not None: cnt += 1
                    cur_W = None
                    cur_b = None
                    source = target
            else:
                for input_node in node.input:
                    if input_node in const_var:
                        if input_node not in added_const:
                            added_const[input_node] = True
                            new_nodes.append(const_var[input_node][1])
                new_nodes.append(node)
                source = node.output[0]
                for old_init in node.input:
                    if old_init in initializers:
                        cur_init = create_initializer_tensor(
                            name=old_init,
                            tensor_array=initializers[old_init],
                            data_type=onnx.TensorProto.FLOAT)
                        new_initializers.append(cur_init)
        elif "merge_vit" in onnx_optimization_flags:
            if node.op_type == "BatchNormalization":
                if onnx_model.graph.node[i - 1].op_type == "Transpose" and onnx_model.graph.node[i + 1].op_type == "Transpose":
                    source_input = onnx_model.graph.node[i - 1].input[0]
                else:
                    source_input = node.input[0]
                bn_node = node
                bn_weight = initializers[bn_node.input[1]]
                bn_bias = initializers[bn_node.input[2]]
                bn_mean = initializers[bn_node.input[3]]
                bn_var = initializers[bn_node.input[4]]
                bn_eps = bn_node.attribute[0].f
                w_bn = np.diag(bn_weight / (np.sqrt(bn_eps + bn_var)))
                b_bn = bn_bias - bn_weight * bn_mean / np.sqrt(bn_var + bn_eps)
                ori_w = w_bn
                ori_b = b_bn

                if i + 2 not in to_merge_linears and i + 2 not in to_merge_others:
                    linear_W_initializer = create_initializer_tensor(
                        name=f"linear{cnt}_W",
                        tensor_array=ori_w,
                        data_type=onnx.TensorProto.FLOAT)

                    linear_b_initializer = create_initializer_tensor(
                        name=f"linear{cnt}_b",
                        tensor_array=ori_b,
                        data_type=onnx.TensorProto.FLOAT
                    )
                    matmul_node = onnx.helper.make_node(
                        name=f'linear{cnt}_MatMul',
                        op_type='MatMul',
                        inputs=[source_input, f'linear{cnt}_W'],
                        outputs=[f'linear{cnt}_intermediate']
                    )
                    add_node = onnx.helper.make_node(
                        name=f'linear{cnt}_Add',
                        op_type='Add',
                        inputs=[f'linear{cnt}_intermediate', f'linear{cnt}_b'],
                        outputs=[node.output[0]]
                    )
                    cnt += 1
                    new_nodes.append(matmul_node)
                    new_nodes.append(add_node)
                    new_initializers.append(linear_W_initializer)
                    new_initializers.append(linear_b_initializer)
            elif i in to_merge_others:
                node.input[0] = source_input
                new_nodes.append(node)
            elif i in to_merge_linears:
                w_cur = initializers[node.input[1]]
                b_cur = initializers[onnx_model.graph.node[i + 1].input[0]]
                W = np.matmul(ori_w, w_cur)
                b = np.matmul(ori_b, w_cur) + b_cur

                linear_W_initializer = create_initializer_tensor(
                    name=f"linear{cnt}_W",
                    tensor_array=W,
                    data_type=onnx.TensorProto.FLOAT)

                linear_b_initializer = create_initializer_tensor(
                    name=f"linear{cnt}_b",
                    tensor_array=b,
                    data_type=onnx.TensorProto.FLOAT
                )
                matmul_node = onnx.helper.make_node(
                    name=f'linear{cnt}_MatMul',
                    op_type='MatMul',
                    inputs=[source_input, f'linear{cnt}_W'],
                    outputs=[f'linear{cnt}_intermediate']
                )
                add_node = onnx.helper.make_node(
                    name=f'linear{cnt}_Add',
                    op_type='Add',
                    inputs=[f'linear{cnt}_intermediate', f'linear{cnt}_b'],
                    outputs=[onnx_model.graph.node[i + 1].output[0]]
                )
                cnt += 1
                new_nodes.append(matmul_node)
                new_nodes.append(add_node)
                new_initializers.append(linear_W_initializer)
                new_initializers.append(linear_b_initializer)
            elif (i - 1 not in to_merge_linears and not (node.op_type == "Transpose"
                    and (onnx_model.graph.node[i - 1].op_type == "BatchNormalization"
                    or onnx_model.graph.node[i + 1].op_type == "BatchNormalization"))):
                new_nodes.append(node)
                source = node.output[0]
                for old_init in node.input:
                    if old_init in initializers:
                        cur_init = create_initializer_tensor(
                            name=old_init,
                            tensor_array=initializers[old_init],
                            data_type=onnx.TensorProto.FLOAT)
                        new_initializers.append(cur_init)
        elif "remove_matmul_inplace" in onnx_optimization_flags:
            transpose_case1 = (node.op_type == "Transpose" and i - 2 >= 0 and
                    onnx_model.graph.node[i - 1].op_type == "MatMul" and onnx_model.graph.node[i - 2].op_type == "Transpose")
            transpose_case2 = (node.op_type == "Transpose" and i - 3 >= 0 and
                    onnx_model.graph.node[i - 1].op_type == "MatMul" and onnx_model.graph.node[i - 3].op_type == "Transpose")
            transpose_case3 = (node.op_type == "Transpose" and i + 2 < len(onnx_model.graph.node) and
                    onnx_model.graph.node[i + 1].op_type == "MatMul" and onnx_model.graph.node[i + 2].op_type == "Transpose")
            transpose_case4 = (node.op_type == "Transpose" and i + 3 < len(onnx_model.graph.node) and
                    onnx_model.graph.node[i + 2].op_type == "MatMul" and onnx_model.graph.node[i + 3].op_type == "Transpose")
            matmul_case1 = (node.op_type == "MatMul" and i - 1 >= 0 and i + 1 < len(onnx_model.graph.node) and
                    onnx_model.graph.node[i - 1].op_type == "Transpose" and onnx_model.graph.node[i + 1].op_type == "Transpose" and
                    onnx_model.graph.node[i - 1].output[0] == node.input[1])
            matmul_case2 = (node.op_type == "MatMul" and i - 2 >= 0 and i + 1 < len(onnx_model.graph.node) and
                    onnx_model.graph.node[i - 2].op_type == "Transpose" and onnx_model.graph.node[i + 1].op_type == "Transpose" and
                    onnx_model.graph.node[i - 2].output[0] == node.input[1])

            if not (transpose_case1 or transpose_case2 or transpose_case3 or transpose_case4):
                if node.op_type == "Constant" and i in list(to_transpose_const.values()):
                    continue
                elif matmul_case1 or matmul_case2:
                    if matmul_case1:
                        source_input = onnx_model.graph.node[i - 1].input[0]
                    else:
                        source_input = onnx_model.graph.node[i - 2].input[0]
                    const_node = onnx_model.graph.node[to_transpose_const[i]]
                    val = nh.to_array(const_node.attribute[0].t).transpose(1, 0)
                    new_tensor = create_initializer_tensor(
                        name=f"Constant{cnt}",
                        tensor_array=val,
                        data_type=onnx.TensorProto.FLOAT
                    )
                    matmul_node = onnx.helper.make_node(
                        name=f'linear{cnt}_MatMul',
                        op_type='MatMul',
                        inputs=[source_input, f"Constant{cnt}"],
                        outputs=[onnx_model.graph.node[i + 1].output[0]]
                    )
                    new_nodes.append(matmul_node)
                    new_initializers.append(new_tensor)
                    cnt += 1
                else:
                    new_nodes.append(node)
                    source = node.output[0]
                    for old_init in node.input:
                        if old_init in initializers:
                            cur_init = create_initializer_tensor(
                                name=old_init,
                                tensor_array=initializers[old_init],
                                data_type=onnx.TensorProto.FLOAT)
                            new_initializers.append(cur_init)
        elif "fix_gtrsb" in onnx_optimization_flags:
            if node.op_type == "Transpose":
                onnx_model.graph.node[i+1].input[0] = node.input[0]
            elif node.op_type == "MatMul" and cnt == 0:
                w_cur = initializers[node.input[1]]
                shape_now = w_cur.shape
                if shape_now[0] == 23328:
                    w_cur = w_cur.reshape(1, 27, 27, 32, -1)
                elif shape_now[0] == 3136:
                    w_cur = w_cur.reshape(1, 7, 7, 64, -1)
                elif shape_now[0] == 1600:
                    w_cur = w_cur.reshape(1, 5, 5, 64, -1)
                else:
                    raise ValueError
                W = w_cur.transpose(0, 3, 1, 2, 4).reshape(-1, shape_now[-1])
                linear_W_initializer = create_initializer_tensor(
                    name=f"matmul{cnt}_W",
                    tensor_array=W,
                    data_type=onnx.TensorProto.FLOAT)

                matmul_node = onnx.helper.make_node(
                    name=f'matmul{cnt}_MatMul',
                    op_type='MatMul',
                    inputs=[node.input[0], f"matmul{cnt}_W"],
                    outputs=[node.output[0]]
                )
                cnt += 1
                new_nodes.append(matmul_node)
                new_initializers.append(linear_W_initializer)
            else:
                for i in range(len(node.input)): node.input[i] = trace(node.input[i])
                new_nodes.append(node)
                source = node.output[0]
                for old_init in node.input:
                    if old_init in initializers and old_init not in [i.name for i in new_initializers]:
                        if "const" in old_init:
                            data_type = onnx.TensorProto.INT64
                        else:
                            data_type = onnx.TensorProto.FLOAT
                        cur_init = create_initializer_tensor(
                            name=old_init,
                            tensor_array=initializers[old_init],
                            data_type=data_type)
                        new_initializers.append(cur_init)
        else:
            # TODO (assigned to Zhuowen): generalize the code to allow the merge of BN + gemm for any general onnx.
            if "merge_bn_reshape_gemm" in onnx_optimization_flags and (i == len(onnx_model.graph.node[started_node:]) - 4) and (node.op_type == "BatchNormalization" and
                    onnx_model.graph.node[-3].op_type == "Constant" and onnx_model.graph.node[-2].op_type == "Reshape" and onnx_model.graph.node[-1].op_type == "Gemm"):
                new_linear_weight, new_linear_bias = fuse_cgan_bn_reshape_gemm(onnx_model, initializers)

                values = nh.to_array(onnx_model.graph.node[-3].attribute[0].t)
                cnode = onnx.helper.make_node(
                    'Constant',
                    inputs=[],
                    outputs=['values1'],
                    value=onnx.helper.make_tensor(
                        name='const_tensor',
                        data_type=onnx.TensorProto.INT64,
                        dims=values.shape,
                        vals=values.flatten().astype(int),
                    ),
                )
                new_nodes.append(cnode)
                rnode = onnx.helper.make_node(
                    'Reshape',
                    inputs=[onnx_model.graph.node[-5].output[0], 'values1'],
                    outputs=['reshape_output'],
                )
                new_nodes.append(rnode)

                linear_W_initializer = create_initializer_tensor(
                    name=f'linear_W1',
                    tensor_array=new_linear_weight,
                    data_type=onnx.TensorProto.FLOAT)
                new_initializers.append(linear_W_initializer)
                linear_b_initializer = create_initializer_tensor(
                    name=f'linear_b1',
                    tensor_array=new_linear_bias,
                    data_type=onnx.TensorProto.FLOAT)
                new_initializers.append(linear_b_initializer)

                cur_node = onnx.helper.make_node(
                    op_type='Gemm',
                    inputs=['reshape_output', 'linear_W1', 'linear_b1'],
                    outputs=['Y'],
                    alpha=1.0,
                    beta=1.0,
                    transB=1
                )
                new_nodes.append(cur_node)
                break

            for i in range(len(node.input)):
                node.input[i] = trace(node.input[i])
            new_nodes.append(node)
            source = node.output[0]
            for old_init in node.input:
                if old_init in initializers:
                    value = initializers[old_init]
                    if value.dtype == 'float32':
                        data_type = onnx.TensorProto.FLOAT
                    elif value.dtype == 'int64':
                        data_type = onnx.TensorProto.INT64
                    else:
                        raise NotImplementedError(value.dtype)
                    cur_init = create_initializer_tensor(
                        name=old_init, tensor_array=value, data_type=data_type)
                    new_initializers.append(cur_init)

    model_outputs = onnx_model.graph.output
    input_node = onnx_model.graph.input[0]


    if 'check_duplicate_upsample_initializers' in onnx_optimization_flags:
        # For cGAN transformers, duplicate initializers appear after using
        # https://github.com/daquexian/onnx-simplifier,
        # which cannot pass onnx.checker.check.
        for i in range(len(new_nodes)):
            if ('Upsample' in new_nodes[i].name
                    and len(new_nodes[i].input) == 2
                    and new_nodes[i].input[1] in initializers):
                for j in range(i + 1, len(new_nodes)):
                    if ('Upsample' in new_nodes[j].name
                            and len(new_nodes[j].input) == 2
                            and new_nodes[i].input[1] == new_nodes[j].input[1]):
                        new_name = f'{new_nodes[j].input[1]}.{j}'
                        assert new_name not in initializers
                        new_init = create_initializer_tensor(
                            name=new_name,
                            tensor_array=initializers[new_nodes[j].input[1]],
                            data_type=onnx.TensorProto.FLOAT)
                        new_nodes[j].input[1] = new_init.name
                        new_initializers.append(new_init)
        new_initializers_dup = new_initializers
        new_initializers = []
        new_initializers_names = []
        for init in new_initializers_dup:
            if init.name not in new_initializers_names:
                new_initializers_names.append(init.name)
                new_initializers.append(init)

    if "remove_squeeze_in_last_layer" in onnx_optimization_flags:
        if new_nodes[-1].op_type == "Squeeze" and len(new_nodes[-1].attribute) == 0:
            new_nodes = new_nodes[:-1]
            model_outputs = [onnx.helper.make_value_info(
                new_nodes[-1].output[0],
                onnx.helper.make_tensor_type_proto(1, [1, 1]))]
            print("Remove the squeeze in the last layer.")

    if "remove_relu_in_last_layer" in onnx_optimization_flags:
        if new_nodes[-1].op_type == "Relu":
            new_nodes = new_nodes[:-1]
            model_outputs[0].name = new_nodes[-1].output[0]
            print("Remove the relu in the last layer.")
    if "fix_gtrsb" in onnx_optimization_flags:
        for i, node in enumerate(new_nodes):
            if node.op_type == "Reshape":
                tmp = node.input[0]
                node.input[0] = new_nodes[i+3].output[0]
                new_nodes[i+1].input[0] = tmp
                new_nodes[i+4].input[0] = node.output[0]
                new_nodes[i] = new_nodes[i+1]
                new_nodes[i+1] = new_nodes[i+2]
                new_nodes[i+2] = new_nodes[i+3]
                new_nodes[i+3] = node
                break
        input_node = onnx.helper.make_tensor_value_info(
            name=onnx_model.graph.input[0].name,
            elem_type=1,
            shape=("unk__195", place_holder[3], place_holder[1], place_holder[2])
        )

    """
    # we should do flatten here, but the collins_yolo_robustness has incompatible onnx version
    # that make the conversion fail. We temporally flatten outside.
    if "flatten_final_output" in onnx_optimization_flags:
        flatten_node = onnx.helper.make_node(
            op_type='Flatten',
            inputs=[new_nodes[-1].output[0]],
            outputs=['Flattened_final_output'],
        )
        new_nodes.append(flatten_node)
        model_outputs[0].name = new_nodes[-1].output[0]
        print("Flattened the final output.")
    """

    new_graph = onnx.helper.make_graph(
        nodes=new_nodes,
        name="CompressedNet",
        inputs=[input_node],
        outputs=model_outputs,
        initializer=new_initializers
    )
    model_def = onnx.helper.make_model(new_graph, producer_name="onnx_example")
    if ('merge_vit' in onnx_optimization_flags
            or 'check_duplicate_upsample_initializers' in onnx_optimization_flags):
        model_def.opset_import[0].version = 9
    else:
        model_def.opset_import[0].version = 13
    model_def = onnx.shape_inference.infer_shapes(model_def)

    onnx.checker.check_model(model_def)
    onnx.save(model_def, save_path)

    if debug:
        errors = np.zeros(100)
        # Workaround for onnx bug, see issue #150
        options = ort.SessionOptions()
        options.inter_op_num_threads = 1
        options.intra_op_num_threads = 1
        for _ in range(100):
            inputs = torch.randn(place_holder).numpy()
            ort_sess = ort.InferenceSession(old_path, sess_options=options)
            input_name = ort_sess.get_inputs()[0].name
            output_name = ort_sess.get_outputs()[0].name
            output1 = ort_sess.run([output_name], {input_name: inputs})

            if "fix_gtrsb" in onnx_optimization_flags:
                inputs = inputs.transpose(0, 3, 1, 2)
            ort_sess = ort.InferenceSession(save_path, sess_options=options)
            input_name = ort_sess.get_inputs()[0].name
            output_name = ort_sess.get_outputs()[0].name
            output2 = ort_sess.run([output_name], {input_name: inputs})

        if "remove_relu_in_last_layer" in onnx_optimization_flags:
            output2 = np.array(output2).clip(min=0)

        errors[_] = np.sum(np.abs(np.array(output1) - np.array(output2)).reshape(-1))

        print(f"Compressed {len(onnx_model.graph.node) - len(model_def.graph.node)} Onnx nodes")
        print('Sum/mean of errors during compression: '
              f'{np.sum(errors)}/{np.mean(errors)}')

    return model_def


if __name__ == '__main__':
    old_path = "../vnncomp2023_benchmarks/benchmarks/cgan/onnx/cGAN_imgSz32_nCh_1.onnx"
    onnx_model = onnx.load(old_path)
    npath = "test.onnx"
    onnx_optimization_flags = ["merge_gemm_reshape_bn", "merge_bn_reshape_gemm"]
    compress_onnx(onnx_model, old_path, npath, onnx_optimization_flags, True)



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
import math
import torch

import arguments


@torch.no_grad()
def input_split_parallel(x_L, x_U, shape=None,
                         cs=None, thresholds=None, split_depth=1, i_idx=None,
                         split_partitions=2):
    """
    Split the x_L and x_U given split_idx and split_depth.
    """
    # FIXME: this function should not be in this file.
    x_L = x_L.flatten(1)
    x_U = x_U.flatten(1)

    x_L_cp = x_L.clone()
    x_U_cp = x_U.clone()

    split_depth = min(split_depth, i_idx.size(1))
    remaining_depth = split_depth
    input_dim = x_L.shape[1]
    while remaining_depth > 0:
        for i in range(min(input_dim, remaining_depth)):
            indices = torch.arange(x_L_cp.shape[0])
            copy_num = x_L_cp.shape[0]//x_L.shape[0]
            idx = i_idx[:,i].repeat(copy_num).long()

            x_L_cp_list, x_U_cp_list = [], []
            for partition in range(split_partitions):
                x_L_cp_tmp = x_L_cp.clone()
                x_U_cp_tmp = x_U_cp.clone()

                lrange = ((partition + 1) * x_L_cp[indices, idx] +
                          (split_partitions - partition - 1) * x_U_cp[indices, idx]) / split_partitions
                urange = (partition * x_L_cp[indices, idx] +
                          (split_partitions - partition) * x_U_cp[indices, idx]) / split_partitions

                x_L_cp_tmp[indices, idx] = lrange
                x_U_cp_tmp[indices, idx] = urange

                x_L_cp_list.append(x_L_cp_tmp)
                x_U_cp_list.append(x_U_cp_tmp)

            x_L_cp = torch.cat(x_L_cp_list)
            x_U_cp = torch.cat(x_U_cp_list)

        remaining_depth -= min(input_dim, remaining_depth)

    split_depth = split_depth - remaining_depth

    new_x_L = x_L_cp.reshape(-1, *shape[1:])
    new_x_U = x_U_cp.reshape(-1, *shape[1:])

    if cs is not None:
        cs_shape = [split_partitions ** split_depth] + [1] * (len(cs.shape) - 1)
        cs = cs.repeat(*cs_shape)
    if thresholds is not None:
        thresholds = thresholds.repeat(split_partitions ** split_depth, 1)
    split_idx = i_idx.repeat(split_partitions ** split_depth, 1)
    return new_x_L, new_x_U, cs, thresholds, split_depth, split_idx


def get_split_depth(x_L, split_partitions=2):
    split_depth = 1
    min_batch_size_ratio = arguments.Config["solver"]["min_batch_size_ratio"]
    batch_size = arguments.Config["solver"]["batch_size"]
    if len(x_L) < min_batch_size_ratio * batch_size:
        min_batch_size = min_batch_size_ratio * batch_size
        split_depth = int(math.log(min_batch_size//len(x_L))//math.log(split_partitions))
        split_depth = max(split_depth, 1)
    return split_depth



#########################################################################
##   This file is part of the α,β-CROWN (alpha-beta-CROWN) verifier    ##
##                                                                     ##
##   Copyright (C) 2021-2024 The α,β-CROWN Team                        ##
##   Primary contacts: Huan Zhang <huan@huan-zhang.com>                ##
##                     Zhouxing Shi <zshi@cs.ucla.edu>                 ##
##                     Kaidi Xu <kx46@drexel.edu>                      ##
##                                                                     ##
##    See CONTRIBUTORS for all author contacts and affiliations.       ##
##                                                                     ##
##     This program is licensed under the BSD 3-Clause License,        ##
##        contained in the LICENCE file in this directory.             ##
##                                                                     ##
#########################################################################
from beta_CROWN_solver import LiRPANet
from attack.attack_pgd import attack_with_general_specs

def use_LiRPANet(model, x, data_min, data_max,
                    list_target_label_arrays,
                    initialization="uniform", GAMA_loss=False):
    wrapped_model = LiRPANet(model, in_size=x.shape)

    model_layers = wrapped_model.net._modules.keys()
    signmerge_layers = []
    for layer_id in model_layers:
        if "/merge" in layer_id:
            signmerge_layers.append(layer_id)

    # For the smallest model, we tend to use the loose approximation
    loose_approx = True if len(signmerge_layers) == 2 else False

    num_attacks = 40
    for _ in range(num_attacks):
        if loose_approx:
            for layer_id in signmerge_layers:
                wrapped_model.net[layer_id].signmergefunction = wrapped_model.net[layer_id].loose_function
        else:
            for layer_id in signmerge_layers:
                wrapped_model.net[layer_id].signmergefunction = wrapped_model.net[layer_id].tight_function

        res, attack_image, attack_margin, all_adv_candidates = attack_with_general_specs(wrapped_model.net, x, data_min, data_max,
                        list_target_label_arrays,
                        initialization, GAMA_loss)

        if res:
            break

        loose_approx = not loose_approx


    return res, attack_image, attack_margin, all_adv_candidates


import json
import pickle
import sys
import time
import traceback
import torch
import yaml
import os

from CTRAIN.verification_systems.abCROWN.complete_verifier.abcrown import ABCROWN
from CTRAIN.verification_systems.abCROWN.complete_verifier.read_vnnlib import read_vnnlib
from CTRAIN.complete_verification.abCROWN.util import get_abcrown_standard_conf

MAX_LOSS = 2**25


def run_abcrown_eval(config, seed, instance, vnnlib_path='../../vnnlib/', model_name='mnist_6_100', model_path='./abCROWN/complete_verifier/models/eran/mnist_6_100_nat.pth', model_onnx_path=None, input_shape=[-1, 1, 28, 28], timeout=600, no_cores=28, par_factor=10):
    print(config, seed, instance)
    std_conf = config
    
    device = 'cuda' if torch.cuda.is_available() else 'mps' if torch.backends.mps.is_available() else 'cpu'
    
    timestamp = time.time()
    
    std_conf['model']['name'] = model_name
    std_conf['model']['path'] = f'/tmp/{model_name}.pth' if model_name is not None else None
    std_conf['model']['onnx_path'] = model_onnx_path if model_onnx_path is not None else None
    std_conf['model']['input_shape'] = input_shape
    
    std_conf['general']['device'] = device
    
    std_conf['bab']['timeout'] = timeout
    
    if not std_conf['solver'].get('mip'):
        std_conf['solver']['mip'] = get_abcrown_standard_conf(timeout=timeout, no_cores=no_cores)['solver']['mip']
    std_conf['solver']['mip']['parallel_solvers'] = no_cores
    
    std_conf['specification']['vnnlib_path_prefix'] = vnnlib_path
    std_conf['specification']['vnnlib_path'] = instance
    std_conf['general']['output_file'] = f'/tmp/out_{timestamp}.pkl'
        
    print(json.dumps(config, indent=2))
    
    with open(f"/tmp/conf_{timestamp}.yaml", "w", encoding='u8') as f:
        yaml.dump(std_conf, f)
    
    abcrown_instance = ABCROWN(
        ['--config', f'/tmp/conf_{timestamp}.yaml']
    )
    
    # Precompile VNN-LIB s.t. each run can access the cache
    _ = read_vnnlib(instance)
    
    start_time = time.time()
    try:
        verification_res = abcrown_instance.main()
    except Exception as e:
        print(type(e), e)
        print(traceback.format_exc())
        return MAX_LOSS, 'unknown'
    end_time = time.time()
    
    os.system(f'rm /tmp/conf_{timestamp}.yaml')
    
    with open(f'/tmp/out_{timestamp}.pkl', 'rb') as f:
        result_dict = pickle.load(f)
    
    result = result_dict['results']
    
    if result == 'unknown':
        print("PENALISING RUNNING TIME DUE TO TIMEOUT!")
        running_time = timeout * par_factor if timeout > (end_time - start_time) else (end_time - start_time) * par_factor
    else:
        running_time = end_time - start_time
    
    return running_time, {"running_time": running_time, "result": result}


if __name__ == "__main__":
    args_path = sys.argv[1]
    result_path = sys.argv[2]
    
    with open(args_path, 'rb') as f:
        args, kwargs = pickle.load(f)
    
    running_time, result = run_abcrown_eval(*args, **kwargs)
    
    with open(result_path, 'wb') as f:
        pickle.dump((running_time, result), f)
    
    sys.exit(0)